/***************************************************************************
 *          ___       ___
 *         /   \     /   \    VESvault
 *         \__ /     \ __/    Encrypt Everything without fear of losing the Key
 *            \\     //                   https://vesvault.com https://ves.host
 *             \\   //
 *     ___      \\_//
 *    /   \     /   \         libVES:                      VESvault API library
 *    \__ /     \ __/
 *       \\     //
 *        \\   //
 *         \\_//              - Key Management and Exchange
 *         /   \              - Item Encryption and Sharing
 *         \___/              - VESrecovery (TM)
 *
 *
 * (c) 2017 - 2022 VESvault Corp
 * Jim Zubov <jz@vesvault.com>
 *
 * GNU General Public License v3
 * You may opt to use, copy, modify, merge, publish, distribute and/or sell
 * copies of the Software, and permit persons to whom the Software is
 * furnished to do so, under the terms of the COPYING file.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 * @title libVES
 * @dev A JavaScript end-to-end encryption interface to VESvault REST API
 * @version 2.16
 *
 * @dev Official source code: https://github.com/vesvault/libVES
 *
 * @author Jim Zubov <jz@vesvault.com> (VESvault Corp)
 *
 ***************************************************************************/
crypto = require('crypto');
crypto.subtle = require('subtle');
XMLHttpRequest = require('xhr2');


if (typeof(libVES) != 'function') function libVES(optns) {
    try {
	if (!crypto.subtle.digest) throw new libVES.Error('Init', 'crypto.subtle is unavailable or improperly implemented');
    } catch (e) {
	if (e instanceof libVES.Error) throw e;
	throw new libVES.Error('Init', 'crypto.subtle is not available');
    }
    for (var k in optns) this[k] = optns[k];
    if (this.domain) this.type = 'secondary';
    else if (this.user) this.type = 'primary';
    else throw new libVES.Error('InvalidValue','Required parameters: user || domain');
    this.unlockedKeys = {};
}

libVES.prototype = {
    constructor: libVES,
    apiUrl: 'https://api.ves.host/v1/',
    pollUrl: 'https://poll.ves.host/v1/',
    wwwUrl: 'https://www.vesvault.com/',
    keyAlgo: 'ECDH',
    keyOptions: {namedCurve: 'P-521'},
    textCipher: 'AES256GCMp',
    defaultHash: 'SHA256',
    propagators: [],
    request: function(method,uri,body,optns) {
	if (!optns) optns = {};
	return new Promise(function(resolve,reject) {
	    var xhr = new XMLHttpRequest();
	    xhr.open(method, (uri.match(/^https\:/) ? uri : this.apiUrl + uri));
	    if (optns.abortFn) optns.abortFn(function() {
		return xhr.abort();
	    });
	    xhr.onreadystatechange = function() {
		switch(xhr.readyState) {
		    case 4:
			if (xhr.response && typeof(xhr.response) == 'object') {
			    if (xhr.response.errors) {
				var errs = xhr.response.errors.map(function(e) {
				    return new libVES.Error(e.type,e.message,e);
				});
				if (errs.length) {
				    if (optns && optns.onerror) try {
					resolve(optns.onerror(errs));
				    } catch (e) {
					reject(e);
				    }
				    else reject(errs[0]);
				}
			    }
			    else resolve(xhr.response.result);
			} else reject(new libVES.Error('BadResponse','Empty response'));
		}
	    };
	    if (body != null) xhr.setRequestHeader('Content-Type','application/json');
	    xhr.setRequestHeader('Accept','application/json');
	    if (this.user && optns.password) xhr.setRequestHeader('Authorization','Basic ' + btoa(this.user + ':' + optns.password));
	    else if (optns.token || this.token) xhr.setRequestHeader('Authorization','Bearer ' + (optns.token ? optns.token : this.token));
	    xhr.responseType = 'json';
	    xhr.send(body);
	}.bind(this));
    },
    get: function(uri,fields,optns) {
	return this.request('GET',this.uriWithFields(uri,fields),null,optns);
    },
    post: function(uri,data,fields,optns) {
	return this.request('POST',this.uriWithFields(uri,fields),JSON.stringify(data),optns);
    },
    uriWithFields: function(uri,fields) {
	return fields ? uri + (uri.match(/\?/) ? '&' : '?') + 'fields=' + this.uriListFields(fields) : uri;
    },
    uriListFields: function(fields) {
	if (typeof(fields) == 'object') {
	    var rs = [];
	    if (fields[0]) rs = fields;
	    else for (var k in fields) {
		if (fields[k]) rs.push(k + (typeof(fields[k]) == 'object' ? '(' + this.uriListFields(fields[k]) + ')' : ''));
	    }
	    return rs.join(',');
	}
	return '';
    },
    elevateAuth: function(optns) {
	var self = this;
	if (optns && (optns.password || optns.token)) return Promise.resolve(optns);
	return (optns && optns.authVaultKey ? Promise.resolve(optns.authVaultKey) : self.getVaultKey()).then(function(vkey) {
	    return vkey.getSessionToken().then(function(tkn) {
		if (!tkn) return optns;
		var o = {token: tkn};
		if (optns) for (var k in optns) o[k] = optns[k];
		return o;
	    });
	});
    },
    login: function(passwd) {
	if (this.token) return this.me();
	var self = this;
	return this.userMe = Promise.resolve(passwd).then(function(passwd) {
	    return self.get('me',{sessionToken: true},{password: passwd}).then(function(data) {
		if (!data.sessionToken) throw new libVES.Error('InvalidValue','Session Token is not received');
		self.token = data.sessionToken;
		return new libVES.User(data,self);
	    });
	});
    },
    logout: function() {
	this.token = this.userMe = this.vaultKey = undefined;
	return this.lock();
    },
    delegate: function(optns) {
	var self = this;
	return libVES.getModule(libVES,'Delegate').then(function(dlg) {
	    return dlg.login(self,null,optns).then(function() {
		return self;
	    });
	});
    },
    getVESflow: function() {
	var self = this;
	if (!this.VESflow) this.VESflow = libVES.getModule(libVES, 'Flow').then(function(fw) {
	    var flow = new fw('VES');
	    var authf = self.authorize.bind(self);
	    self.authorize = function(msg) {
		flow.setValue(msg);
		return authf(msg);
	    };
	    var outf = self.logout.bind(self);
	    self.logout = function() {
		flow.logout();
		return outf();
	    };
	    return flow;
	});
	return this.VESflow;
    },
    flow: function(start, optns) {
	var self = this;
	return self.getVESflow().then(function(flow) {
	    return flow.get().then(function(auth) {
		return self.authorize(auth);
	    }).catch(function(e) {
		if (!start || (e && e.code == 'Reload')) throw e;
		var url = self.wwwUrl + 'vv/unlock?url=' + encodeURIComponent(document.location.href) + '&domain=' + encodeURIComponent(self.domain);
		if (optns) for (var k in optns) if (optns[k] != null) url += '&' + encodeURIComponent(k) + '=' + encodeURIComponent(optns[k]);
		flow.url = undefined;
		return flow.setValue(undefined).then(function() {
		    return flow.addToken(url).then(function(url) {
			document.location.replace(url);
			throw new libVES.Error('Redirect', 'Starting VES Authorization...');
		    });
		});
	    });
	});
    },
    authorize: function(msg) {
	var self = this;
	return this.logout().then(function() {
	    if (msg.token) self.token = msg.token;
	    if (msg.domain) self.domain = msg.domain;
	    if (msg.externalId) self.externalId = msg.externalId;
	    return (msg.VESkey ? self.unlock(msg.VESkey) : self.me()).then(function() {
		return self;
	    });
	});
    },
    carry: function(optns) {
	var cr = {};
	if (this.externalId) cr.externalId = this.externalId;
	if (this.domain) cr.domain = this.domain;
	if (this.token) cr.token = this.token;
	if (this.VESkey) cr.VESkey = this.VESkey;
	try {
	    sessionStorage['libVES_carry'] = JSON.stringify(cr);
	} catch (e) {
	    return Promise.reject(e);
	}
	return Promise.resolve(true);
    },
    pick: function(optns) {
	try {
	    if (!sessionStorage['libVES_carry']) throw new libVES.Error('NotFound','No libVES_carry in sessionStorage');
	    var cr = JSON.parse(sessionStorage['libVES_carry']);
	    delete(sessionStorage['libVES_carry']);
	    return this.authorize(cr);
	} catch (e) {
	    return Promise.reject(e);
	}
    },
    me: function() {
	var self = this;
	if (!this.userMe) this.userMe = this.get('me').then((function(data) {
	    return new libVES.User(data,self);
	}).bind(this));
	return this.userMe;
    },
    unlock: function(veskey) {
	var self = this;
	return this.getVaultKey().then(function(vkey) {
	    return vkey.unlock(Promise.resolve(veskey)).then(function(cryptoKey) {
		if (self.VESkey === null) self.VESkey = veskey;
		if (!self.token && self.type == 'secondary') return vkey.getSessionToken().then(function(tkn) {
		    self.token = tkn;
		    return cryptoKey;
		});
		return cryptoKey;
	    });
	}).then(function(ck) {
	    return self.handleAttn().catch(function() {}).then(function() {
		return ck;
	    });
	});
    },
    lock: function() {
	this.unlockedKeys = {};
	return Promise.resolve(true);
    },
    reset: function(val) {
	this.userMe = undefined;
	return this.lock().then(function() {
	    return val;
	});
    },
    getVaultKey: function() {
	var self = this;
	switch (this.type) {
	    case 'primary': return this.me().then(function(me) {
		return me.getCurrentVaultKey();
	    });
	    case 'secondary': return (this.vaultKey || (this.vaultKey = this.prepareExternals({externalId: self.externalId}).then(function(ext) {
		var vKey = new libVES.VaultKey({type: 'secondary', externals: ext},self);
		return vKey.getField('encSessionToken').then(function(tk) {
		    return vKey;
		});
	    })));
	    default: throw new libVES.Error('Internal','Invalid libVES.type: ' + this.type);
	}
    },
    getShadowKey: function() {
	return this.me().then(function(me) {
	    return me.getShadowVaultKey();
	});
    },
    getVaultKeysById: function() {
	return this.me().then(function(me) {
	    return me.getVaultKeys().then(function(vaultKeys) {
		return Promise.all(vaultKeys.map(function(e,i) {
		    return e.getId();
		})).then(function(ids) {
		    var rs = {};
		    for (var i = 0; i < ids.length; i++) rs[ids[i]] = vaultKeys[i];
		    return rs;
		});
	    });
	});
    },
    getItems: function() {
	var self = this;
	return this.getVaultKey().then(function(k) {
	    return k.getId().then(function(kid) {
		return k.getVaultEntries({type: true, deleted: true, file: {creator: true, externals: true}, vaultKey: {type: true, user: true}}).then(function(ves) {
		    var vis = {};
		    var vlst = [];
		    for (var i = 0; i < ves.length; i++) {
			var viid = ves[i].vaultItem.id;
			if (!vis[viid]) {
			    var vi = ves[i].vaultItem;
			    if (!vi.file) vi.file = undefined;
			    if (!vi.vaultKey) vi.vaultKey = undefined;
			    vi = vis[viid] = self.getItem(vi);
			    vlst.push(vi);
			    vi.vaultEntryByKey[kid] = ves[i];
			}
		    }
		    return vlst;
		});
	    });
	});
    },
    getItem: function(data) {
	return new libVES.VaultItem(data,this);
    },
    postItem: function(data) {
	var vi = new libVES.VaultItem(data,this);
	return vi.validate().then(function() {
	    return vi.post();
	});
    },
    usersToKeys: function(users) {
	var self = this;
	return Promise.all(users.map(function(u) {
	    if (typeof(u) == 'object') {
		if (u instanceof libVES.VaultKey) return [u];
		else if (u instanceof libVES.External) return [new libVES.VaultKey({externals:[u]},self)];
		else if (u instanceof libVES.User) return self.getUserKeys(u);
		else if (u instanceof Array || u.domain != null || u.externalId != null) return self._matchSecondaryKey(u, u.user, u.appUrl).then(function(vkey) {
		    return [vkey];
		});
	    }
	    return self.getUserKeys(self._matchUser(u));
	})).then(function(ks) {
	    var rs = [];
	    for (var i = 0; i < ks.length; i++) for (var j = 0; j < ks[i].length; j++) rs.push(ks[i][j]);
	    return rs;
	});
    },
    _matchUser: function(u) {
	if (typeof(u) == 'object') {
	    if (u instanceof libVES.User) return u;
	    else return new libVES.User(u,this);
	} else if (typeof(u) == 'string' && u.match(/^\S+\@\S+$/)) return new libVES.User({email: u},this);
	throw new libVES.Error('BadUser',"Cannot match user: " + u,{value: u});
    },
    _matchSecondaryKey: function(ext, user, appUrl) {
	var self = this;
	var m = function() {
	    return libVES.getModule(libVES.Domain,ext.domain).catch(function(e) {
		return {
		    userToVaultRef: function(u) {
			return u.getEmail().then(function(email) {
			    return email.toLowerCase();
			});
		    },
		    vaultRefToUser: function(ext) {
			var email = ext.externalId ? ext.externalId.replace(/\!.*/, '') : '';
			if (!email.match(/\@/)) throw new libVES.Error('NotFound', 'externalId is not an email for non-existing vault');
			return new libVES.User({email: email}, self);
		    }
		};
	    });
	};
	return (ext.externalId ? self.prepareExternals(ext) : m().then(function(dom) {
	    return Promise.resolve(user || self.me()).then(function(u) {
		return dom.userToVaultRef(u);
	    }).then(function(ex) {
		return self.prepareExternals([ex]);
	    });
	}).catch(function(e) {
	    throw new libVES.Error('NotFound', 'Cannot match externalId for domain:' + ext.domain + ', user:' + user + '. Define libVES.Domain.' + ext.domain + '.userToVaultRef(user) to return a valid reference.',{error: e});
	})).then(function(exts) {
	    var vkey = new libVES.VaultKey({externals: exts, creator: self.me()},self);
	    return vkey.getId().then(function() {
		return vkey;
	    }).catch(function(e) {
		if (e.code != 'NotFound') throw e;
		return Promise.resolve(user || m().then(function(dom) {
		    return dom.vaultRefToUser(exts[0]);
		})).catch(function(e) {
		    throw new libVES.Error('NotFound', 'No matching secondary vault key',{error: e});
		}).then(function(u) {
		    return self.me().then(function(me) {
			return Promise.all([me.getId(),u.getId()]).then(function(ids) {
			    if (ids[0] == ids[1]) return self.getSecondaryKey(exts,true);
			}).catch(function(e) {
			    if (e.code != 'NotFound') throw e;
			}).then(function(rs) {
			    return rs || self.createTempKey(self._matchUser(u)).then(function(vkey) {
				return vkey.setField('externals',exts).then(function() {
				    if (appUrl) vkey.setField('appUrl', appUrl);
				    return vkey;
				});
			    });
			});
		    });
		});
	    });
	});
    },
    getUserKeys: function(usr) {
	var self = this;
	return usr.getActiveVaultKeys().catch(function(e) {
	    if (e.code == 'NotFound') return [];
	    throw e;
	}).then(function(keys) {
	    return Promise.all(keys.map(function(k,i) {
		return k.getPublicCryptoKey().then(function() {
		    return k;
		}).catch(function() {});
	    })).then(function(keys) {
		var rs = [];
		for (var i = 0; i < keys.length; i++) if (keys[i]) rs.push(keys[i]);
		return rs;
	    });
	}).then(function(keys) {
	    if (!keys.length) return self.createTempKey(usr).then(function(k) {
		return [k];
	    });
	    return keys;
	});
    },
    createTempKey: function(usr, optns) {
	var self = this;
	var key = new libVES.VaultKey({type: 'temp', algo: this.keyAlgo, user: usr}, self);
	var veskey = this.generateVESkey(usr);
	return key.generate(veskey, optns).then(function(k) {
	    key.setField('vaultItems', veskey.then(function(v) {
		var vi = new libVES.VaultItem({type: 'password'}, self);
		return usr.getActiveVaultKeys().then(function() {
		    return [self.me(), self.getVaultKey(), usr];
		}).catch(function(e) {
		    if (e.code != 'NotFound') throw e;
		    var sh = self.type == 'secondary' ? [self.me(), self.getVaultKey()] : [self.me()];
		    if (self.propagators) for (var i = 0; i < self.propagators.length; i++) sh.push(self.propagators[i]);
		    return sh;
		}).then(function(sh) {
		    return Promise.all(sh);
		}).then(function(sh) {
		    return vi.shareWith(sh, v, false).then(function() {
			return [vi];
		    });
		});
	    }));
	    key.setField('creator', self.me());
	    return key;
	});
    },
    generateVESkey: function(usr) {
	var buf = new Uint8Array(24);
	crypto.getRandomValues(buf);
	return Promise.resolve(libVES.Util.ByteArrayToB64(buf));
    },
    setVESkey: function(veskey,lost,options) {
	var self = this;
	return this.me().then(function(me) {
	    return (new libVES.VaultKey({type: 'current', algo: self.keyAlgo, user: me},self)).generate(Promise.resolve(veskey),options).then(function(k) {
		return me.getCurrentVaultKey().then(function(cur) {
		    return (cur ? cur.unlock().then(function() {
			return k.rekeyFrom(cur);
		    }).catch(function(e) {
			if (!lost) throw e;
		    }) : Promise.resolve(null)).then(function() {
			var r;
			if (cur && lost) r = cur.setField('type','lost').then(function() {
			    k.user = undefined;
			    return me.setField('vaultKeys',[cur,k]).then(function() {
				return me;
			    });
			});
			else r = k;
			me.currentVaultKey = me.activeVaultKeys = me.shadowVaultKey = undefined;
			if (!cur || !lost) me.vaultKeys = undefined;
			return r;
		    });
		}).then(function(r) {
		    return self.elevateAuth(options).then(function(optns) {
			return r.post(undefined, undefined, optns);
		    });
		}).catch(function(e) {
		    self.reset();
		    throw e;
		}).then(function(post) {
		    return self.reset(post);
		}).then(function() {
		    return self.getVaultKey();
		});
	    });
	});
    },
    prepareExternals: function(ext) {
	var self = this;
	if (!ext) return Promise.reject(new libVES.Error('InvalidValue','External reference is required'));
	return Promise.resolve(ext).then(function(ext) {
	    if (!(ext instanceof Array)) ext = [ext];
	    if (ext.length < 1) throw new libVES.Error('InvalidValue','External reference is required');
	    var rs = [];
	    for (var i = 0; i < ext.length; i++) {
		rs[i] = (typeof(ext[i]) == 'object') ? {externalId: ext[i].externalId, domain: ext[i].domain} : {externalId: ext[i]};
		if (!rs[i].domain && !(rs[i].domain = self.domain)) throw new libVES.Error('InvalidValue','External reference: domain is required');
		if (!rs[i].externalId) throw new libVES.Error('InvalidValue','External reference: externalId is required');
	    }
	    return rs;
	});
    },
    getSecondaryKey: function(ext, force) {
	var self = this;
	return this.prepareExternals(Promise.resolve(ext).then(function(e) {
	    if (e.domain && !e.externalId) return libVES.getModule(libVES.Domain, e.domain).then(function(mod) {
		return self.me().then(function(me) {
		    return mod.userToVaultRef(me, self);
		});
	    });
	    return e;
	})).then(function(ext) {
	    var vkey = new libVES.VaultKey({externals: ext}, self);
	    return vkey.getId().then(function(id) {
		return vkey;
	    }).catch(function(e) {
		if (!force || e.code != 'NotFound') throw e;
		return self.setSecondaryKey(ext);
	    });
	});
    },
    setSecondaryKey: function(ext,veskey,optns) {
	var self = this;
	return this.prepareExternals(ext).then(function(ext) {
	    if (!veskey) veskey = self.generateVESkey();
	    return self.me().then(function(me) {
		return (new libVES.VaultKey({type: 'secondary', algo: self.keyAlgo, user: me, externals: ext},self)).generate(veskey,optns).then(function(k) {
		    return self.getSecondaryKey(ext).then(function(k2) {
			return k.rekeyFrom(k2);
		    }).catch(function(e) {
			delete(k.vaultEntries);
			return k;
		    });
		}).then(function(k) {
		    var vi = new libVES.VaultItem({type: "password"},self);
		    k.setField('vaultItems',[vi]);
		    return Promise.resolve(veskey).then(function(v) {
			if (!v) throw new libVES.Error('InvalidValue','VESkey cannot be empty');
			return vi.shareWith([me],v,false).then(function() {
			    return self.elevateAuth(optns);
			}).then(function(optns) {
			    return k.post(undefined, undefined, optns).then(function(post) {
				return k.setField('id', post.id, false).then(function() {
				    k.fieldUpdate = {id: true};
				    return k;
				});
			    });
			});
		    });
		});
	    });
	});
    },
    setAnonymousKey: function(veskey, optns) {
	var self = this;
	return this.prepareExternals({externalId: self.externalId}).then(function(ext) {
	    if (!veskey) throw new libVES.Error('InvalidValue','VESkey cannot be empty');
	    return (new libVES.VaultKey({type: 'secondary', algo: (optns && optns.algo ? optns.algo : self.keyAlgo), externals: ext},self)).generate(veskey, optns).then(function(k) {
		return k.post(undefined, ['encSessionToken'], optns).then(function(post) {
		    return k.setField('id', post.id, false).then(function() {
			k.fieldUpdate = {id: true};
		    });
		}).catch(function(e) {
		    if (!e || e.code != 'Unauthorized') throw e;
		}).then(function() {
		    return self.unlock(veskey);
		});
	    });
	});
    },
    setShadow: function(usrs,optns) {
	var self = this;
	if (!(usrs instanceof Array)) return Promise.reject(new libVES.Error('InvalidValue', 'usrs must be an array'));
	if (!usrs.length) {
	    return self.getShadowKey().then(function(sh) {
		return (sh ? sh.getId().then(function(id) {
		    return self.elevateAuth(optns).then(function(optns) {
			return self.post('vaultKeys/' + id, {'$op': 'delete'}, undefined, optns). then(function() {
			    return null;
			});
		    });
		}) : null);
	    });
	}
	if (!optns || !optns.n) return Promise.reject(new libVES.Error('InvalidValue','optns.n must be an integer'));
	var rkey = new Uint8Array(32);
	crypto.getRandomValues(rkey);
	var algo = optns.v ? libVES.Scramble.algo[optns.v] : libVES.Scramble.RDX;
	if (!algo) return Promise.reject(new libVES.Error('InvalidValue','Unknown scramble algorithm: ' + optns.v));
	var s = new algo(optns.n);
	return s.explode(rkey,usrs.length,optns).then(function(tkns) {
	    return self.me().then(function(me) {
		me.activeVaultKeys = undefined;
		return me.setField('shadowVaultKey',new libVES.VaultKey({type: 'shadow', user: me, algo: self.keyAlgo},self).generate(rkey,optns),false).then(function(k) {
		    return me.getCurrentVaultKey().then(function(curr) {
			return k.rekeyFrom(curr).catch(function() {}).then(function() {
			    libVES.Object._refs = {"#/":k};
			    k.setField('vaultItems',Promise.all(tkns.map(function(tk,i) {
				var vi = new  libVES.VaultItem({type: 'secret'},self);
				return vi.shareWith([usrs[i]],tk,false).then(function() {
				    return vi;
				});
			    })).then(function(vis) {
				delete(libVES.Object._refs);
				return vis;
			    }));
			    return self.elevateAuth(optns).then(function(optns) {
				return k.post(undefined, undefined, optns);
			    });
			});
		    });
		}).catch(function(e) {
		    me.shadowVaultKey = undefined;
		    throw e;
		}).then(function() {
		    me.currentVaultKey = me.shadowVaultKey = me.activeVaultKeys = undefined;
		    return me.getShadowVaultKey();
		});
	    });
	});
    },
    getFile: function(fileRef) {
	var self = this;
	return self.prepareExternals(fileRef).then(function(ext) {
	    return new libVES.File({externals: ext},self);
	});
    },
    getFileItem: function(fileRef) {
	var self = this;
	return self.getFile(fileRef).then(function(file) {
	    return new libVES.VaultItem({file: file},self);
	});
    },
    getValue: function(fileRef) {
	return this.getFileItem(fileRef).then(function(vaultItem) {
	    return vaultItem.get();
	});
    },
    putValue: function(fileRef,value,shareWith) {
	var self = this;
	return this.getFileItem(fileRef).then(function(vaultItem) {
	    return vaultItem.setField('type',libVES.VaultItem.Type._detect(value)).then(function() {
	        return Promise.resolve(shareWith || self.getFileItem(fileRef).then(function(vi) {
		    return vi.getShareList();
		}).catch(function(e) {
		    return self.usersToKeys([{domain: VES.domain, externalId: VES.externalId}]);
		})).then(function(shareWith) {
		    return vaultItem.shareWith(shareWith,value);
		});
	    });
	});
    },
    shareFile: function(fileRef,shareWith) {
	return this.getFileItem(fileRef).then(function(vaultItem) {
	    return vaultItem.shareWith(shareWith);
	});
    },
    fileExists: function(fileRef) {
	return this.getFileItem(fileRef).then(function(vaultItem) {
	    return vaultItem.getId().then(function(id) {
		return true;
	    }).catch(function(e) {
		if (e.code == 'NotFound') return false;
		throw e;
	    });
	});
    },
    deleteFile: function(fileRef) {
	return this.getFileItem(fileRef).then(function(item) {
	    return item.delete();
	});
    },
    newSecret: function(cls) {
	if (!cls) cls = this.textCipher;
	else cls = cls.split('.')[0];
	return libVES.getModule(libVES.Cipher,cls).then(function(ci) {
	    return (new ci()).getSecret().then(function(buf) {
		return cls + '.' + libVES.Util.ByteArrayToB64W(buf);
	    });
	});
    },
    secretToCipher: function(secret) {
	var ss = secret.split('.');
	return libVES.getModule(libVES.Cipher,ss[0]).then(function(cls) {
	    return new cls(libVES.Util.B64ToByteArray(ss[1]));
	});
    },
    encryptText: function(openText,secret) {
	return this.secretToCipher(secret).then(function(ci) {
	    return ci.encrypt(libVES.Util.StringToByteArray(openText),true).then(function(buf) {
		return libVES.Util.ByteArrayToB64W(buf);
	    });
	});
    },
    decryptText: function(cipherText,secret) {
	return this.secretToCipher(secret).then(function(ci) {
	    return ci.decrypt(libVES.Util.B64ToByteArray(cipherText),true).then(function(buf) {
		return libVES.Util.ByteArrayToString(buf);
	    });
	});
    },
    hashText: function(text,cls) {
	if (cls) cls = cls.split('.')[0];
	else cls = this.defaultHash;
	return libVES.getModule(libVES.Util,['Hash',cls]).then(function(mod) {
	    return mod.hash(libVES.Util.StringToByteArray(text)).then(function(buf) {
		return cls + '.' + libVES.Util.ByteArrayToB64W(buf);
	    });
	});
    },
    found: function(veskeys,vaultKeys) {
	var self = this;
	return Promise.resolve(veskeys).then(function(veskeys) {
	    var chain = Promise.resolve(0);
	    if (veskeys && !(veskeys instanceof Array)) veskeys = [veskeys];
	    return (vaultKeys ? Promise.resolve(vaultKeys) : self.me().then(function(me) {
		var rs = [];
		return me.getVaultKeys().then(function(vaultKeys) {
		    return Promise.all(vaultKeys.map(function(vaultKey,i) {
			return vaultKey.getType().then(function(t) {
			    switch (t) {
				case 'temp': case 'lost': case 'recovery': rs.push(vaultKey);
			    }
			});
		    }));
		}).then(function() {
		    return rs;
		});
	    })).then(function(vaultKeys) {
		if (!(vaultKeys instanceof Array)) vaultKeys = [vaultKeys];
		return Promise.all(vaultKeys.map(function(vaultKey,i) {
		    return vaultKey.getRecovery().then(function(rcv) {
			return rcv.unlock();
		    }).catch(function(e) {
			var rs = vaultKey.unlock();
			if (veskeys) veskeys.map(function(veskey,i) {
			    rs = rs.catch(function() {
				return vaultKey.unlock(veskey);
			    });
			});
			return rs;
		    }).then(function() {
			chain = chain.then(function(ct) {
			    return vaultKey.rekey().then(function() {
				return ct + 1;
			    });
			}).catch(function(e) {console.log(e);});;
		    }).catch(function() {});
		}));
	    }).then(function() {
		return chain;
	    });
	}).then(function(ct) {
	    if (ct) return ct + self.found();
	});
    },
    getMyRecoveries: function() {
	var self = this;
	return self.me().then(function(me) {
	    return me.getVaultKeys().then(function(vaultKeys) {
		return Promise.all(vaultKeys.map(function(e,i) {
		    return e.getType();
		})).then(function(types) {
		    var rs = [];
		    for (var i = 0; i < types.length; i++) switch (types[i]) {
			case 'recovery': case 'shadow':
			    rs.push(vaultKeys[i].getRecovery());
		    }
		    return Promise.all(rs);
		});
	    });
	});
    },
    getFriendsRecoveries: function() {
	var self = this;
	return self.me().then(function(me) {
	    return me.getFriendsKeyItems().then(function(vaultItems) {
		var vaultKeys = [];
		return Promise.all(vaultItems.map(function(e,i) {
		    return e.getVaultKey().then(function(vk) {
			vaultKeys[i] = vk;
			if (vk) return vk.getType();
		    });
		})).then(function(types) {
		    var rs = [];
		    for (var i = 0; i < types.length; i++) switch (types[i]) {
			case 'recovery': case 'shadow':
			    rs.push(vaultKeys[i].getRecovery([vaultItems[i]]));
		    }
		    return Promise.all(rs);
		});
	    });
	});
    },
    getUnlockableKeys: function() {
	var self = this;
	return self.getVaultKey().then(function(vk) {
	    return vk.getField('unlockableVaultKeys').then(function(vks) {
		var rs = {};
		return Promise.all(vks.map(function(e, i) {
		    return e.getId().then(function(id) {
			rs[id] = e;
		    });
		})).then(function() {
		    return rs;
		});
	    });
	});
    },
    getAttn: function() {
	var self = this;
	var rs = {};
	return self.get('attn').then(function(attn) {
	    return attn ? Promise.all([(attn.vaultKeys ? Promise.all(attn.vaultKeys.map(function(vk, i) {
		return new libVES.VaultKey(vk, self);
	    })).then(function(vks) {
		rs.vaultKeys = vks;
	    }) : null)]) : null;
	}).then(function() {
	    return rs;
	});
    },
    handleAttn: function(attn) {
	var self = this;
	return (attn ? Promise.resolve(attn) : self.getAttn()).then(function(attn) {
	    if (attn) return Promise.all([(attn.vaultKeys ? self.attnVaultKeys(attn.vaultKeys) : null)]);
	});
    },
    attnVaultKeys: function(vkeys) {
	var self = this;
	return Promise.all(vkeys.map(function(vkey, i) {
	    console.log('attnVaultKeys:', vkey);
	    return vkey.getType().then(function(type) {
		switch (type) {
		    case 'temp':
		    case 'lost':
			return vkey.getUser().then(function(user) {
			    return self.me().then(function(me) {
				return Promise.all([user.getId(), me.getId()]).then(function(ids) {
				    if (ids[0] == ids[1]) return type == 'temp' ? vkey.rekey()
				    : self.elevateAuth({authVaultKey: vkey}).then(function(optns) {
					return vkey.rekey(optns);
				    });
				    else return vkey.getVaultItems().then(function(vis) {
					return self.elevateAuth().then(function(optns) {
					    return Promise.all(vis.map(function(vi, i) {
						return vi.reshareWith([user], undefined, optns);
					    }));
					});
				    });
				});
			    });
			});
		    case 'recovery':
			return vkey.getRecovery().then(function(rcv) {
			    return rcv.recover();
			});
		}
	    });
	}));
    },
    setKeyAlgo: function(optns) {
	if (typeof(optns) == 'string') {
	    this.keyOptions = null;
	    return this.keyAlgo = optns;
	}
	var algo = libVES.Algo.fromKeyOptions(optns);
	if (algo) {
	    this.keyOptions = optns;
	    return this.keyAlgo = algo;
	}
    }
};

libVES.Error = function(code, msg, optns) {
    if (libVES.Error[code] && (libVES.Error[code].prototype instanceof libVES.Error)) return new (libVES.Error[code])(msg, optns);
    this.code = code;
    this.init(msg, optns);
};

libVES.Error.prototype.init = function(msg, optns) {
    this.message = msg;
    if (optns) for (var k in optns) this[k] = optns[k];
};

libVES.Error.prototype.toString = function() {
    return this.message || this.code;
};

libVES.Error.NotFound = function(msg, optns) {
    this.init(msg, optns);
};
libVES.Error.NotFound.prototype = new libVES.Error('NotFound');

libVES.Error.InvalidValue = function(msg, optns) {
    this.init(msg, optns);
};
libVES.Error.InvalidValue.prototype = new libVES.Error('InvalidValue');

libVES.Error.InvalidKey = function(msg, optns) {
    this.init(msg, optns);
};
libVES.Error.InvalidKey.prototype = new libVES.Error('InvalidKey');

libVES.Error.Redirect = function(msg, optns) {
    this.init(msg, optns);
};
libVES.Error.Redirect.prototype = new libVES.Error('Redirect');

libVES.Error.Internal = function(msg, optns) {
    this.init(msg, optns);
};
libVES.Error.Internal.prototype = new libVES.Error('Internal');


libVES.getModule = function(sectn,mods) {
    var mod;
    if (mods instanceof Array) mod = mods[0];
    else mods = [mod = mods];
    if (sectn[mod]) return mods.length > 1 ? libVES.getModule(sectn[mod],mods.slice(1)) : Promise.resolve(sectn[mod]);
    if (sectn.loadModule) {
	if (sectn.loadModule[mod]) return sectn.loadModule[mod];
    } else sectn.loadModule = {};
    return sectn.loadModule[mod] = libVES.loadModule(sectn,mod).then(function(m) {
	delete(sectn.loadModule[mod]);
	sectn[mod] = m;
	return ((mods instanceof Array) && mods.length > 1 ? libVES.getModule(m,mods.slice(1)) : m);
    });
};
libVES.getModuleFunc = function(sectn,mod,then) {
    return function() { var m = libVES.getModule(sectn,mod); return then ? m.then(then) : m; };
};
libVES.loadModule = function(sectn,mod) {
    return Promise.reject(new libVES.Error('Internal',"Cannot load " + sectn + '.' + mod));
};
libVES.maxKeyLen = 48896;
libVES.maxEncDataLen = 32768;


if (!libVES.Domain) libVES.Domain = {};

libVES.Util = {
    B64ToByteArray: function(s) {
	var buf = new Uint8Array(s.length);
	var boffs = 0;
	for (var i = 0; i < s.length; i++) {
	    var p = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/-_".indexOf(s[i]);
	    if (p >= 0) {
		if (p >= 64) p -= 2;
		buf[boffs >> 3] |= p << 2 >> (boffs & 7);
		boffs += 6;
		if ((boffs & 7) < 6) buf[boffs >> 3] |= p << (8 - (boffs & 7));
	    }
	}
	var l = boffs >> 3;
	var buf2 = new Uint8Array(l);
	for (var i = 0; i < l; i++) buf2[i] = buf[i];
	return buf2.buffer;
    },
    ByteArrayToB64D: function(b,dict) {
	var buf = new Uint8Array(b);
	var s = "";
	var boffs = 0;
	while ((boffs >> 3) < buf.byteLength) {
	    var c = (buf[boffs >> 3] << (boffs & 7)) & 0xfc;
	    boffs += 6;
	    if (((boffs & 7) < 6) && ((boffs >> 3) < buf.byteLength)) c |= (buf[boffs >> 3] >> (6 - (boffs & 7)));
	    s += dict[c >> 2];
	}
	for (; boffs & 7; boffs += 6) s += dict.substr(64);
	return s;
    },
    ByteArrayToB64: function(b) {
	return libVES.Util.ByteArrayToB64D(b,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=");
    },
    ByteArrayToB64W: function(b) {
	return libVES.Util.ByteArrayToB64D(b,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_");
    },
    StringToByteArray: function(s) {
	if ((s instanceof ArrayBuffer) || (s instanceof Uint8Array)) return s;
	var rs = new Uint8Array(4 * s.length);
	var j = 0;
	for (var i = 0; i < s.length;i++) {
	    var c = s.charCodeAt(i);
	    if (c >= 0x80) {
		if (c >= 0x0800) {
		    if (c >= 0x10000) {
			rs[j++] = (c >> 16) | 0xf0;
			rs[j++] = ((c >> 12) & 0x3f) | 0x80;
		    } else rs[j++] = ((c >> 12) & 0x0f) | 0xe0;
		    rs[j++] = ((c >> 6) & 0x3f) | 0x80;
		} else rs[j++] = ((c >> 6) & 0x1f) | 0xc0;
		rs[j++] = (c & 0x3f) | 0x80;
	    } else rs[j++] = c;
	}
	return rs.slice(0,j).buffer;
    },
    ByteArrayToString: function(b) {
	var buf = new Uint8Array(b);
	var rs = '';
	var c;
	for (var i = 0; i < buf.length; i++) {
	    var v = buf[i];
	    if (v & 0x80) {
		if (v & 0x40) {
		    c = ((v & 0x1f) << 6) | (buf[++i] & 0x3f);
		    if (v & 0x20) {
			c = (c << 6) | (buf[++i] & 0x3f);
			if (v & 0x10) c = ((c & 0xffff) << 6) | (buf[++i] & 0x3f);
		    }
		} else c = -1;
	    } else c = buf[i];
	    rs += String.fromCharCode(c);
	}
	return rs;
    },
    loadWasm: function(src) {
	return new Promise(function(resolve, reject) {
	    var sc = document.createElement('script');
	    sc.async = false;
	    sc.src = src;
	    sc.onload = resolve;
	    sc.onerror = reject;
	    document.getElementsByTagName('head')[0].appendChild(sc);
	});
    },
    PEM: {
	toDER: function(pem) {
	    var pp = pem.match(/-----BEGIN.*?-----\s*\r?\n([A-Za-z0-9\/\+\=\s\r\n]*)-----END/);
	    if (!pp) throw new libVES.Error('Internal','PEM formatted key expected');
	    return new Uint8Array(libVES.Util.B64ToByteArray(pp[1]));
	},
	decode: function(pem) {
	    return libVES.Util.ASN1.decode(libVES.Util.PEM.toDER(pem));
	},
	import: function(pem,optns) {
	    return libVES.Util.ASN1.import(libVES.Util.PEM.toDER(pem),optns);
	},
	fromDER: function(der) {
	},
	encode: function(der,sgn) {
	    return '-----BEGIN ' + sgn + '-----\r\n' + libVES.Util.ByteArrayToB64(der).match(/.{1,64}/g).join("\r\n") + '\r\n-----END ' + sgn + '-----';
	}
    },
    ASN1: {
	decode: function(der,fStruct) {
	    var p = 0;
	    var data = function() {
		var l = der[p++];
		var len;
		if (l < 128) len = l;
		else {
		    len = 0;
		    for (var i = 128; i < l; i++) len = (len << 8) | der[p++];
		}
		if (p + len > der.length) throw new libVES.Error('Internal',"Invalid ASN.1 package");
		return der.slice(p,p = p + len);
	    };
	    var rs = [];
	    for (; p < der.length;) {
		if (!fStruct && p) {
		    rs.push(der.slice(p));
		    break;
		}
		var tag = der[p++];
		switch (tag) {
		    case 48:
			rs.push(libVES.Util.ASN1.decode(data(),true));
			break;
		    case 6:
			rs.push(new libVES.Util.OID(data()));
			break;
		    case 2:
			var d = data();
			var v = 0;
			for (var i = 0; i < d.length; i++) v = (v << 8) | d[i];
			rs.push(v);
			break;
		    case 5:
			data();
			rs.push(null);
			break;
		    default:
			rs.push(data());
			break;
		}
	    }
	    return rs;
	},
	encode: function(data,fStruct) {
	    var i2a = function(v) {
		if (v < 0) throw new libVES.Error('Internal',"Negative value for ASN.1 integer!");
		var rs = [];
		do {
		    rs.push(v & 0xff);
		    v >>= 8;
		} while (v > 0);
		return rs.reverse();
	    };
	    var bufs = [];
	    var buf = function(tag,bf) {
		var b = new Uint8Array(bf);
		var l = b.length;
		var rs;
		if (l <= 127) {
		    rs = new Uint8Array(l + 2);
		    rs[1] = l; 
		    rs.set(b,2);
		} else {
		    var lb = i2a(l);
		    rs = new Uint8Array(l + lb.length + 2);
		    rs[1] = 128 + lb.length;
		    rs.set(lb,2);
		    rs.set(b,2 + lb.length);
		}
		rs[0] = tag;
		bufs.push(rs);
		return rs;
	    };
	    var d;
	    for (var i = 0; i < data.length; i++) if (fStruct || !i) switch (typeof(d = data[i])) {
		case 'object':
		    if (d == null) buf(5,new Uint8Array(0));
		    else if (d instanceof Array) buf(48,libVES.Util.ASN1.encode(d,true));
		    else if (d instanceof libVES.Util.OID) buf(6,d.getBuffer());
		    else if (d instanceof Uint8Array || d instanceof ArrayBuffer) buf((d.ASN1type || 4),d);
		    else throw new libVES.Error('Internal',"ASN.1 encode - Unknown type");
		    break;
		case 'number':
		    buf(2,i2a(d));
		    break;
		default: throw new libVES.Error('Internal',"ASN.1 encode - Unknown type");
	    } else bufs.push(d);
	    var l = 0;
	    for (var i = 0; i < bufs.length; i++) l += bufs[i].length;
	    var der = new Uint8Array(l);
	    var p = 0;
	    for (i = 0, p = 0; i < bufs.length; p += bufs[i].length, i++) der.set(bufs[i],p);
	    return der;
	},
	import: function(der,optns) {
	    var k = optns && optns.decoded || libVES.Util.ASN1.decode(der)[0];
	    if (!k) throw new libVES.Error('Internal','Empty ASN.1 package?');
	    var i = 0;
	    if (typeof(k[i]) == 'number') (optns || (optns = {})).version = k[i++];
	    if (typeof(k[i]) == 'object' && (k[i][0] instanceof libVES.Util.OID)) return k[i][0].object().then(function(m) {
		return m.import(k[i][1],function(call,optns) {
		    return new Promise(function(resolve,reject) {
			switch (call) {
			    case 'container': return resolve(der);
			    default: return resolve(k[i + 1]);
			}
		    });
		},optns);
	    });
	    return libVES.Util.ASN1.import(libVES.Util.ASN1.encode([[0,[new libVES.Util.OID('1.2.840.113549.1.1.1'),null],der]])).catch(function(e) {
		throw new libVES.Error('Internal',"Unknown key format",{error: e});
	    });
	},
	setType: function(t,buf) {
	    var rs = new Uint8Array(buf);
	    rs.ASN1type = t;
	    return rs;
	}
    },
    OID: function(s) {
	if (s instanceof Uint8Array) {
	    var rs = [ Math.floor(s[0] / 40), s[0] % 40 ];
	    var r = 0;
	    for (var p = 1; p < s.length; p++) {
		var v = s[p];
		r = (r << 7) | (v & 0x7f);
		if (!(v & 0x80)) {
		    rs.push(r);
		    r = 0;
		}
	    }
	    this.value = rs.join('.');
	} else this.value = s;
    },
    Key: {
	fromDER: function(der, callbk) {
	    var asn = libVES.Util.ASN1.decode(der)[0];
	    if (typeof(asn[0]) == 'number') {
		var asn2 = libVES.Util.ASN1.decode(asn[2])[0];
		var pub = libVES.Util.ASN1.decode(asn2[2]);
		console.log('pub', pub);
		if (pub) return callbk(pub[0].slice(1), asn2[1], pub[0][0]);
		else callbk(null, asn2[1]);
	    } else return callbk(asn[1].slice(1), null, asn[1][0]);
	},
	pubASN1: function(pub, optns) {
	    var p = new Uint8Array(pub.byteLength + 1);
	    p[0] = optns && optns.pubBits ? optns.pubBits : 0;
	    p.set(new Uint8Array(pub), 1);
	    p.ASN1type = 3;
	    return p;
	},
	toPKCS: function(pub, priv, asn1hdr, optns) {
	    if (priv) {
		var seq = [1, priv];
		if (pub) {
		    p = libVES.Util.ASN1.encode([libVES.Util.Key.pubASN1(pub, optns)]);
		    p.ASN1type = 0xa1;
		    seq.push(p);
		}
		var buf = libVES.Util.ASN1.encode([seq]);
		return libVES.Util.ASN1.encode([[0, asn1hdr, buf]]);
	    }
	    return libVES.Util.ASN1.encode([[asn1hdr, libVES.Util.Key.pubASN1(pub, optns)]]);
	},
	export: function(pub, priv, asn1hdr, optns) {
	    var pkcs = libVES.Util.Key.toPKCS(pub, priv, asn1hdr, optns);
	    if (priv) return libVES.Util.PKCS8.encode8(pkcs, optns);
	    return libVES.Util.PEM.encode(pkcs, 'PUBLIC KEY');
	}
    },
    PKCS1: {
	import: function(args,chain,optns) {
	    return chain('container',optns).then(function(der) {
		return crypto.subtle.importKey('spki',der,{name:'RSA-OAEP', hash:'SHA-1'},true,['encrypt']).catch(function(e) {
		    return crypto.subtle.importKey('pkcs8',der,{name:'RSA-OAEP', hash:'SHA-1'},true,['decrypt']);
		});
	    });
	},
	encode: function(key,optns) {
	    return crypto.subtle.exportKey('spki',key).then(function(der) {
		return libVES.Util.PEM.encode(der,'PUBLIC KEY');
	    });
	}
    },
    PKCS5: {
	import: function(args,chain,optns) {
	    var f = chain;
	    for (var i = args.length - 1; i >= 0; i--) f = (function(obj,fp) {
		if (obj[0] instanceof libVES.Util.OID) return function(call,optns) {
		    return obj[0].object().then(function(m) {
			return m[call](obj[1],fp,optns);
		    });
		};
		else return fp;
	    })(args[i],f);
	    return f('import',optns).then(function(der) {
		return libVES.Util.ASN1.import(new Uint8Array(der));
	    }).catch(function(e) {
		throw new libVES.Error('InvalidKey',"Cannot import the private key (Invalid VESkey?)");
	    });
	},
	export: function(chain,optns) {
	    var args = [];
	    var f = chain;
	    if (!optns || !(optns.members instanceof Array)) throw new libVES.Error('Internal','PKCS#5: optns.members must be an array');
	    for (var i = optns.members.length - 1; i >= 0; i--) f = (function(obj,fp,idx) {
		return function(call,optns) {
		    return obj[call](fp,optns).then(function(v) {
			if (call == 'export') args[idx] = v;
			return v;
		    });
		};
	    })(optns.members[i],f,i);
	    return f('export',optns).then(function() {
		return [new libVES.Util.OID('1.2.840.113549.1.5.13'), args];
	    });
	}
    },
    PKCS8: {
	encode: function(key, optns) {
	    return crypto.subtle.exportKey('pkcs8',key).catch(function(e) {
		console.log('PKCS8 failed, trying JWK...');
		return crypto.subtle.exportKey('jwk', key).then(function(jwk) {
		    return libVES.Util.PKCS8.fromJWK(jwk);
		});
	    }).then(function(pkcs8) {
		return libVES.Util.PKCS8.encode8(pkcs8, optns);
	    });
	},
	encode8: function(pkcs8, optns) {
	    var ops = {};
	    for (var k in optns) ops[k] = optns[k];
	    if (!ops.members) ops.members = [
		libVES.getModule(libVES.Util,'PBKDF2'),
		libVES.getModule(libVES.Cipher,'AES256CBC')
	    ];
	    return Promise.all(ops.members).then(function(ms) {
		ops.members = ms;
		ops.content = pkcs8;
		var rec = [];
		if (ops.password) return libVES.Util.PKCS5.export(function(call,optns) {
		    rec[1] = optns.content;
		    return Promise.resolve();
		},ops).then(function(data) {
		    rec[0] = data;
		    return libVES.Util.PEM.encode(libVES.Util.ASN1.encode([rec]),'ENCRYPTED PRIVATE KEY');
		});
		else if (ops.opentext) return libVES.Util.PEM.encode(pkcs8,'PRIVATE KEY');
		else throw new libVES.Error('Internal','No password for key export (opentext=true to export without password?)');
	    });
	},
	fromJWK: function(jwk) {
	    if (jwk.kty != 'EC') throw new libVES.Error('Internal', 'kty=="EC" expected (workaround for Firefox)');
	    var pubx = libVES.Util.B64ToByteArray(jwk.x);
	    var puby = libVES.Util.B64ToByteArray(jwk.y);
	    var pub = new Uint8Array(pubx.byteLength + puby.byteLength + 2);
	    pub[0] = 0;
	    pub[1] = 4;
	    pub.set(new Uint8Array(pubx), 2);
	    pub.set(new Uint8Array(puby), pubx.byteLength + 2);
	    pub.ASN1type = 3;
	    pub = libVES.Util.ASN1.encode([pub]);
	    pub.ASN1type = 0xa1;
	    return libVES.Util.ASN1.encode([[
		0,
		[
		    new libVES.Util.OID('1.2.840.10045.2.1'),
		    new libVES.Util.OID((function(crvs) {
			for (var k in crvs) if (crvs[k] == jwk.crv) return k;
			throw new libVES.Error('Internal', 'Unknown named curve: ' + jwk.crv);
		    })(libVES.Util.EC.namedCurves))
		],
		libVES.Util.ASN1.encode([[
		    1,
		    libVES.Util.B64ToByteArray(jwk.d),
		    pub
		]])
	    ]]).buffer;
	},
	toJWK: function(key) {
	    var der = libVES.Util.ASN1.decode(key)[0];
	    var idx = 0;
	    if (!(der[0] instanceof Array)) idx++;
	    switch (String(der[idx][0])) {
		case '1.2.840.10045.2.1':
		case '1.3.132.1.12':
		case '1.3.132.112': // Firefox on Mac - apparently a misspelling of the former
		    break;
		default:
		    console.log(der);
		    throw new libVES.Error('Internal', 'EC PKCS8 expected');
	    }
	    var der2 = idx ? libVES.Util.ASN1.decode(der[idx + 1])[0] : null;
	    var pub = idx ? (der2[2] ? libVES.Util.ASN1.decode(der2[2])[0] : null) : der[1];
	    var publ = 0;
	    if (pub && pub.byteLength > 2 && pub[1] == 4) publ = pub.byteLength / 2 - 1;
	    return {
		crv: libVES.Util.EC.namedCurves[der[idx][1]],
		d: (idx ? libVES.Util.ByteArrayToB64W(der2[1]) : undefined),
		ext: true,
		key_ops: (idx ? ['deriveKey', 'deriveBits'] : []),
		kty: 'EC',
		x: (publ ? libVES.Util.ByteArrayToB64W(pub.slice(2, publ + 2)) : ''),
		y: (publ ? libVES.Util.ByteArrayToB64W(pub.slice(publ + 2, 2 * publ + 2)) : '')
	    };
	}
    },
    PBKDF2: {
	deriveKey: function(args, pwd, algo) {
	    return crypto.subtle.importKey('raw', libVES.Util.StringToByteArray(pwd), 'PBKDF2', false, ['deriveKey']).then(function(k) {
		var keyargs = {name:'PBKDF2', salt:args[0], iterations:args[1]};
		var ka;
		if (args[2] && (args[2][0] instanceof libVES.Util.OID)) ka = args[2][0].object().then(function(obj) {
		    obj.setArgs(keyargs, args[2][1]);
		    return keyargs;
		});
		else {
		    keyargs.hash = 'SHA-1';
		    ka = Promise.resolve(keyargs);
		}
		return ka.then(function(keyargs) {
		    return crypto.subtle.deriveKey(keyargs, k, algo, true, ['encrypt', 'decrypt']);
		});
	    });
	},
	import: function(args,chain,optns) {
	    if (!optns || !optns.password) throw new libVES.Error('InvalidKey',"VESkey is not supplied");
	    var pwd = (typeof(optns.password) == 'function') ? optns.password() : optns.password;
	    return chain('info').then(function(info) {
		return libVES.Util.PBKDF2.deriveKey(args,pwd,info.algorithm).then(function(k) {
		    return chain('import',{key: k});
		});
	    });
	},
	export: function(chain,optns) {
	    if (!optns || !optns.password) throw new libVES.Error('InvalidKey',"VESkey is not supplied");
	    var pwd = (typeof(optns.password) == 'function') ? optns.password() : optns.password;
	    var args = [new Uint8Array(8),((optns && optns.KDF && optns.KDF.iterations) || 2048)];
	    crypto.getRandomValues(args[0]);
	    return chain('info').then(function(info) {
		return libVES.Util.PBKDF2.deriveKey(args,pwd,info.algorithm).then(function(k) {
		    optns.key = k;
		    return chain('export',optns).then(function() {
			return [new libVES.Util.OID('1.2.840.113549.1.5.12'), args];
		    });
		});
	    });
	}
    },
    EC: {
	import: function(args,chain,optns) {
	    return chain('container',optns).then(function(der) {
		var oid = String(args);
		var curve = libVES.Util.EC.namedCurves[oid] || oid;
		var a = {name:'ECDH',namedCurve: curve};
		return crypto.subtle.importKey('spki',der,a,true,[]).catch(function() {
		    return crypto.subtle.importKey('pkcs8',der,a,true,['deriveKey','deriveBits']);
		}).catch(function(e) {
		    console.log('PKCS8 failed, trying JWK...');
		    var jwk = libVES.Util.PKCS8.toJWK(der);
		    return crypto.subtle.importKey('jwk', jwk, a, true, jwk.key_ops);
		}).catch(function(e) {
		    console.log('trying WasmECDH...');
		    return libVES.Algo.ECDH.importWasm(curve, der);
		});
	    });
	},
	namedCurves: {
	    "1.3.132.0.10": "P-256",
	    "1.3.132.0.34": "P-384",
	    "1.3.132.0.35": "P-521"
	}
    },
    Hash: {
	SHA1: {
	    setArgs: function(args, optns) {
		args.hash = 'SHA-1';
		return args;
	    },
	    hash: function(buf) {
		return crypto.subtle.digest('SHA-1',buf);
	    }
	},
	SHA256: {
	    setArgs: function(args, optns) {
		args.hash = 'SHA-256';
		return args;
	    },
	    hash: function(buf) {
		return crypto.subtle.digest('SHA-256',buf);
	    }
	},
	SHA384: {
	    setArgs: function(args, optns) {
		args.hash = 'SHA-384';
		return args;
	    },
	    hash: function(buf) {
		return crypto.subtle.digest('SHA-384',buf);
	    }
	},
	SHA512: {
	    setArgs: function(args, optns) {
		args.hash = 'SHA-512';
		return args;
	    },
	    hash: function(buf) {
		return crypto.subtle.digest('SHA-512',buf);
	    }
	}
    }
};

libVES.Util.OID.prototype = {
    object: function() {
	var o = libVES.Util.OID[this.value];
	if (!o) throw new libVES.Error('Internal',"Unknown object identifier: " + this.value);
	return o();
    },
    getBuffer: function() {
	var oid = this.value.split(/\./).map(function(v) { return Number(v); });
	var rs = [oid[0] * 40 + oid[1]];
	for (var i = 2; i < oid.length; i++) {
	    var n = oid[i];
	    var v = [n & 0x7f];
	    n >>= 7;
	    while (n) {
		v.push((n & 0x7f) | 0x80);
		n >>= 7;
	    }
	    for (var j = v.length - 1; j >= 0; j--) rs.push(v[j]);
	}
	return new Uint8Array(rs);
    },
    toString: function() {
	return this.value;
    }
}

libVES.Util.OID['1.2.840.113549.1.5.13'] = libVES.getModuleFunc(libVES, ['Util', 'PKCS5']);
libVES.Util.OID['1.2.840.113549.1.5.12'] = libVES.getModuleFunc(libVES, ['Util', 'PBKDF2']);
libVES.Util.OID['2.16.840.1.101.3.4.1.42'] = libVES.getModuleFunc(libVES, ['Cipher', 'AES256CBC']);
libVES.Util.OID['1.2.840.113549.1.1.1'] = libVES.getModuleFunc(libVES, ['Util', 'PKCS1']);
libVES.Util.OID['1.2.840.10045.2.1'] = libVES.Util.OID['1.3.132.1.12'] = libVES.Util.OID['1.3.132.112'] = libVES.getModuleFunc(libVES, ['Util', 'EC']);
libVES.Util.OID['1.2.840.113549.2.7'] = libVES.getModuleFunc(libVES, ['Util', 'Hash', 'SHA1']);
libVES.Util.OID['1.2.840.113549.2.9'] = libVES.getModuleFunc(libVES, ['Util', 'Hash', 'SHA256']);
libVES.Util.OID['1.2.840.113549.2.10'] = libVES.getModuleFunc(libVES, ['Util', 'Hash', 'SHA384']);
libVES.Util.OID['1.2.840.113549.2.11'] = libVES.getModuleFunc(libVES, ['Util', 'Hash', 'SHA512']);
libVES.Util.OID['1.3.6.1.4.1.53675.3.5'] = libVES.getModuleFunc(libVES, ['Algo', 'OQS', 'Util']);

libVES.Math = {
    pad: function(a,n) {
	if (isNaN(n)) n = 0;
	if (typeof(a) == 'number') {
	    var rs = [0];
	    for (var i = 0;;) {
		var f = (rs[i] = a & 0xff) & 0x80;
		a >>= 8;
		if (++i >= n && a == (f ? -1 : 0)) break;
	    }
	    return new Uint8Array(rs);
	}
	if (a.length >= n) return a;
	var rs = new Uint8Array(n);
	rs.set(a,0);
	var sgn = (a.length > 0 && (a[a.length - 1] & 0x80)) ? 0xff : 0;
	rs.fill(sgn,a.length);
	return rs;
    },
    add: function(a,b) {
	var l = (a.length > b.length ? a.length : b.length) + 1;
	a = this.pad(a,l);
	b = this.pad(b,l);
	var rs = new Uint8Array(l);
	var c = 0;
	var hi = 0;
	var hs = 0;
	for (var i = 0; i < l; i++) {
	    var v = a[i] + b[i] + c;
	    if ((rs[i] = v & 0xff) != (hs ? 0xff : 0)) hi = i;
	    hs = v & 0x80;
	    c = v >> 8;
	}
	return rs.slice(0,hi + 1);
    },
    sub: function(a,b) {
	return this.add(a,this.neg(b));
    },
    neg: function(a) {
	if (typeof(a) == 'number') return -a;
	var rs = new Uint8Array(a.length + 1);
	var c = 1;
	var o = 0;
	for (var i = 0; i < a.length; i++) {
	    var v = (~a[i] & 0xff) + c;
	    o = c & (v == 0x80);
	    rs[i] = v & 0xff;
	    c = v >> 8;
	}
	if (o) rs[i++] = 0;
	return rs.slice(0,i);
    },
    cmp: function(a,b) {
	var l = a.length > b.length ? a.length : b.length;
	a = this.pad(a,l);
	b = this.pad(b,l);
	var f = true;
	for (var i = l - 1; i >= 0; i--) {
	    if (f) {
		if ((a[i] ^ b[i]) & 0x80) return ((a[i] & 0x80) ? -1 : 1);
		f = false;
	    }
	    if (a[i] > b[i]) return 1;
	    if (a[i] < b[i]) return -1;
	}
	return 0;
    },
    isNeg: function(a) {
	if (typeof(a) == 'number') return a < 0;
	return a.length && (a[a.length - 1] & 0x80);
    },
    isZero: function(a) {
	if (typeof(a) == 'number') return a == 0;
	for (var i = 0; i < a.length; i++) if (a[i]) return false;
	return true;
    },
    mul: function(a,b) {
	var neg = false;
	if (this.isNeg(a)) {
	    neg = true;
	    a = this.neg(a);
	} else a = new Uint8Array(a);
	var int = typeof(b) == 'number';
	if (this.isNeg(b)) {
	    neg = !neg;
	    b = this.neg(b);
	} else if (!int) b = new Uint8Array(b);
	var sh_a = new libVES.Math.Shifter(a);
	if (!int) var sh_b = new libVES.Math.Shifter(b);
	var rs = this.pad(0,1);
	while (true) {
	    if ((int ? b : b[0]) & 0x01) rs = this.add(rs,a);
	    b = int ? b >> 1 : sh_b.shr();
	    if (this.isZero(b)) break;
	    a = sh_a.shl();
	}
	return neg ? this.neg(rs) : rs;
    },
    div_qr: function(a,b) {
	var neg = false;
	if (this.isNeg(a)) {
	    neg = true;
	    a = this.neg(a);
	} else a = new Uint8Array(a);
	if (typeof(b) == 'number') b = this.pad(b);
	if (this.isNeg(b)) {
	    neg = !neg;
	    b = this.neg(b);
	} else b = new Uint8Array(b);
	for (var l = b.length; l > 0 && !b[l - 1]; l--);
	if (!l) return null;
	var sh_b = new libVES.Math.Shifter(b);
	var q;
	for (var sh = (a.length - l + 1) * 8; sh >= 0; sh--) {
	    var d = sh_b.get(sh);
	    if (this.cmp(a,d) >= 0) {
		a = this.sub(a,d);
		if (!q) {
		    q = new Uint8Array(((sh + 1) >> 3) + 1);
		    q.fill(0);
		}
		q[sh >> 3] |= 1 << (sh & 7);
	    }
	}
	if (!q) q = new Uint8Array([0]);
	return {q: neg ? this.neg(q) : q, r: neg ? this.neg(a) : a};
    },
    div: function(a,b) {
	return this.div_qr(a,b).q;
    },
    mulv: function(v1,v2) {
	if (v1.length != v2.length) throw new libVES.Error('Internal','mulv: vectors have different size');
	var rs = this.pad(0);
	for (var i = 0; i < v1.length; i++) rs = this.add(rs,this.mul(v1[i],v2[i]));
	return rs;
    },
    matrixReduce: function(matrix) {
	var m = [];
	for (var i = 0; i < matrix.length; i++) {
	    m[i] = [];
	    for (var j = 0; j < matrix[i].length; j++) m[i][j] = matrix[i][j];
	}
	for (var i = 0; i < m.length; i++) {
	    var q = m[i][i];
	    for (var ii = 0; ii < m.length; ii++) if (ii != i) {
		var p = m[ii][i];
		for (var j = 0; j < m[i].length; j++) m[ii][j] = this.sub(this.mul(m[ii][j],q),this.mul(m[i][j],p));
	    }
	}
	return m;
    },
    
    hexChars: '0123456789abcdef',
    hex: function(a) {
	var rs = [];
	for (var i = 0; i < a.length; i++) rs[a.length - i - 1] = this.hexChars[a[i] >> 4] + this.hexChars[a[i] & 0x0f];
	return rs.join('');
    },
    Shifter: function(a) {
	this.shifts = [a];
	this.offs = 0;
    }
};
libVES.Math.Shifter.prototype = {
    get: function(offs) {
	var self = this;
	var sh = offs & 7;
	var bs = offs >> 3;
	var shf;
	var v = (shf = function(sh) {
	    if (!self.shifts[sh]) {
		if (sh <= 0) throw new libVES.Error('Internal','Shifter offset is not available');
		var prev = shf(sh - 1);
		self.shifts[sh] = libVES.Math.add(prev,prev);
	    }
	    return self.shifts[sh];
	})(sh);
	if (bs <= 0) return v.length + bs >= 1 ? v.slice(-bs) : libVES.Math.pad(v[v.length - 1] & 0x80 ? -1 : 1);
	var rs = new Uint8Array(v.length + bs);
	rs.fill(0,0,bs);
	rs.set(v,bs);
	return rs;
    },
    shl: function() {
	return this.get(++this.offs);
    },
    shr: function() {
	return this.get(--this.offs);
    }
};

libVES.Object = function(data) {
    for (var k in data) this[k] = data[k];
    if (typeof(Trigger) != 'undefined') this.trigger = Trigger.resolve(this);
};

libVES.Object.prototype = {
    fieldList: {id: true},
    fieldExtra: {},
    fieldClass: {},
    fieldSets: [],
    init: function(data,VES,refs) {
	this.VES = VES;
	this.fieldUpdate = data.id ? {id: true} : {};
	this.setFields(data,data.id == null);
	if (refs) for (var k in refs) this[k] = Promise.resolve(refs[k]);
    },
    setFields: function(data,up) {
	var self = this;
	var chg = false;
	for (var k in data) {
	    if (up === undefined || up) this.fieldUpdate[k] = true;
	    if (this[k] instanceof Promise) {
		this[k] = undefined;
		chg = true;
	    }
	    if (this[k] === undefined) this[k] = Promise.resolve(data[k]).then((function(k) {
		return function(v) {
		    var clsf;
		    if (self.fieldClass[k]) return (clsf = function(v) {
			if (v instanceof libVES.Object) return v;
			else if (v instanceof Array) return v.map(function(vv) { return clsf(vv); });
			else if (v) return new (self.fieldClass[k])(v,self.VES);
		    })(v);
		    return v;
		};
	    })(k));
	    else return Promise.reject(new libVES.Error('Internal',"Unknown field: " + k));
	}
	if (chg && self.trigger) self.trigger.resolve(self);
	return Promise.resolve(self);
    },
    setField: function(fld,val,upd) {
	var flds = {};
	flds[fld] = val;
	return this.setFields(flds,upd).then(function(self) {
	    return self[fld];
	});
    },
    getFields: function(fldlst, force) {
	var self = this;
	var flds = {};
	for (var i = 0; i < this.fieldSets.length; i++) {
	    for (var k in this.fieldSets[i]) if (fldlst[k]) {
		for (var k in this.fieldSets[i]) if ((force || !this[k]) && !(flds[k] instanceof Object)) flds[k] = this.fieldSets[i][k];
		break;
	    }
	}
	for (var k in fldlst) if ((force || !this[k]) && !(flds[k] instanceof Object)) flds[k] = fldlst[k];
	for (var k in flds) {
	    this.loadFields(flds);
	    break;
	}
	var plst = [];
	for (var k in fldlst) {
	    plst.push(Promise.resolve(this[k]).then(function(v) {
		if (!(v instanceof libVES.Object)) return v;
		return v.getFields(fldlst[k], force);
	    }));
	}
	var rs = {};
	return Promise.all(plst).then(function(lst) {
	    var i = 0;
	    for (var k in fldlst) rs[k] = lst[i++];
	    return rs;
	});
    },
    getField: function(fld,fldlst,force) {
	var self = this;
	if (!this[fld] || force) {
	    var flds = {};
	    for (var i = 0; i < this.fieldSets.length; i++) if (this.fieldSets[i][fld]) {
		for (var k in this.fieldSets[i]) flds[k] = this.fieldSets[i][k];
		break;
	    }
	    if (fldlst) flds[fld] = fldlst;
	    if (!flds[fld]) {
		var cls = self.fieldClass[fld];
		flds[fld] = cls ? cls.prototype.fieldList : true;
	    }
	    this.loadFields(flds,force);
	}
	return this[fld];
    },
    loadFields: function(flds,force,optns) {
	var self = this;
	var req = this.id ? this.id.then(function(id) {  return self.VES.get(self.apiUri + '/' + id,flds,optns); }) : self.postData().then(function(data) {
	    data['$op'] = 'fetch';
	    return self.VES.post(self.apiUri,data,flds,optns);
	}).then(function(data) {
	    if (data.id) {
		self.id = Promise.resolve(data.id);
		self.fieldUpdate = {id: true};
	    }
	    return data;
	});
	for (var k in flds) if (force || this[k] === undefined) {
	    this[k] = req.then((function(fld) {
		var cls = self.fieldClass[fld];
		return function(data) {
		    if (cls && data[fld]) return ((data[fld] instanceof Array) ? data[fld].map(function(v) {
			return new cls(v,self.VES);
		    }) : new cls(data[fld],self.VES));
		    return data[fld];
		};
	    })(k));
	}
    },
    reset: function() {
	for (var k in this.fieldClass) delete(this[k]);
	return Promise.resolve();
    },
    getId: function() {
	return this.id ? Promise.resolve(this.id) : this.getField('id');
    },
    postData: function(fields,refs,parent) {
	if (refs && parent) for (var k in refs) if (refs[k] === this) return Promise.resolve({"$ref": k});
	var data = {};
	var prs = [];
	var self = this;
	var fmt = function(v,a) {
	    if (v instanceof libVES.Object) return v.postData(a,refs,self);
	    else if (v instanceof Array) return Promise.all(v.map(function(vv,i) {
		return fmt(vv,a);
	    }));
	    else return v;
	};
	var pf = function(k,pr,a) {
	    if (!(pr instanceof Promise)) pr = fmt(pr,a);
	    if (pr instanceof Promise) prs.push(pr.then(function(pr2) {
		return Promise.resolve(fmt(pr2,a)).then(function(v) {
		    data[k] = v;
		});
	    }));
	    else data[k] = pr;
	};
	if (!(fields instanceof Object)) fields = this.fieldUpdate;
	if (fields) for (var k in fields) if (this[k] !== undefined) pf(k,this[k],fields[k]);
	return Promise.all(prs).then(function() {
	    return data;
	});
    },
    post: function(fields,rfields,optns) {
	var self = this;
	if (!optns) optns = {};
	if (optns.retry == null) optns.retry = 3;
	return this.postData(fields,optns.refs).then(function(d) {
	    var op = {
		onerror: function(errors) {
		    if (optns.retry-- <= 0) throw new libVES.Error('RequestFailed',"Retry count exceeded",{errors: errors});
		    var rs = [];
		    for (var i = 0; i < errors.length; i++) {
			if (!errors[i].path) throw errors[i];
			rs.push(self.resolveErrorPath(errors[i]));
		    }
		    return Promise.all(rs).then(function() {
			return self.post(fields,rfields,optns);
		    });
		}
	    };
	    for (var k in optns) op[k] = optns[k];
	    return self.VES.post(self.apiUri,d,rfields,op);
	});
    },
    resolveErrorPath: function(e,idx) {
	var self = this;
	if (!e.path) throw e;
	if (!idx) idx = 0;
	if (e.path.length == idx) return this.resolveError(e,null);
	var f = e.path[idx++];
	if (this[f] === undefined) throw new libVES.Error('BadPath',"Path not found: " + f,{error: e});
	return Promise.resolve(this[f]).then(function(v) {
	    if (v instanceof libVES.Object) return v.resolveErrorPath(e,idx);
	    else if ((e.path.length > idx) && (v instanceof Array)) {
		var i = e.path[idx++];
		if (v[i] === undefined) throw new libVES.Error('BadPath',"Path not found: " + i,{error: e});
		else if (v[i] instanceof libVES.Object) return v[i].resolveErrorPath(e,idx);
		else throw e;
	    } else return self.resolveError(e,f);
	});
    },
    resolveError: function(e,field) {
	throw e;
    }
};


libVES.User = function(data,VES,refs) {
    this.init(data,VES,refs);
};

libVES.VaultKey = function(data,VES,refs) {
    this.init(data,VES,refs);
};

libVES.VaultItem = function(data,VES,refs) {
    this.vaultEntryByKey = {};
    this.init(data,VES,refs);
};

libVES.External = function(data,VES,refs) {
    this.init(data,VES,refs);
};

libVES.File = function(data,VES,refs) {
    this.init(data,VES,refs);
};

libVES.Event = function(data, VES, refs) {
    this.init(data, VES, refs);
};

libVES.Session = function(data, VES, refs) {
    this.init(data, VES, refs);
};

libVES.oldDomain = libVES.Domain;
libVES.Domain = function(dom, VES) {
    this.init({id: (dom ? dom : VES.domain)}, VES);
};
(function(old) {
    if (old) for (var k in old) libVES.Domain[k] = old[k];
})(libVES.oldDomain);

libVES.User.prototype = new libVES.Object({
    apiUri: 'users',
    fieldList: {id: true, email: true, type: true, firstName: true, lastName: true},
    fieldExtra: {vaultKeys: true, activeVaultKeys: true, currentVaultKey: true},
    fieldClass: {vaultKeys: libVES.VaultKey, activeVaultKeys: libVES.VaultKey, currentVaultKey: libVES.VaultKey, shadowVaultKey: libVES.VaultKey, friendsKeyItems: libVES.VaultItem},
    getEmail: function() {
	return this.getField('email');
    },
    getFirstName: function() {
	return this.getField('firstName');
    },
    getLastName: function() {
	return this.getField('lastName');
    },
    getFullName: function() {
	var self = this;
	return this.getFirstName().then(function(f) {
	    return self.getLastName().then(function(l) {
		return f ? (l ? f + ' ' + l : f) : l;
	    });
	});
    },
    getVaultKeys: function() {
	return this.getField('vaultKeys');
    },
    getActiveVaultKeys: function() {
	var self = this;
	if (!this.activeVaultKeys && (this.currentVaultKey || this.shadowVaultKey)) return this.getCurrentVaultKey().then(function(curr) {
	    return curr ? self.getShadowVaultKey().then(function(sh) {
		return sh ? [curr,sh] : [curr];
	    }) : [];
	});
	return this.getField('activeVaultKeys');
    },
    getFriendsKeyItems: function() {
	return this.getField('friendsKeyItems');
    },
    getCurrentVaultKey: function() {
	return this.getField('currentVaultKey');
    },
    getShadowVaultKey: function() {
	return this.getField('shadowVaultKey');
    },
    getExternals: function() {
	return this.getField('externals');
    },
    getExternalsByDomain: function() {
	return this.getExternals().then(function(ex) {
	    var rs = {};
	    for (var i = 0; i < ex.length; i++) (rs[ex[i].domain] || (rs[ex[i].domain] = [])).push(ex[i]);
	    return rs;
	});
    },
    unlock: function(veskey) {
	return this.getCurrentVaultKey().then(function(k) {
	    return k.unlock(veskey);
	});
    },
    lock: function(veskey) {
	if (this.currentVaultKey) return this.currentVaultKey.then(function(k) {
	    return k.lock();
	});
    },
});


libVES.VaultKey.prototype = new libVES.Object({
    apiUri: 'vaultKeys',
    fieldList: {id: true, algo: true, type: true, publicKey: true, privateKey: true},
    fieldClass: {user: libVES.User, vaultItems: libVES.VaultItem, externals: libVES.External, unlockableVaultKeys: libVES.VaultKey, creator: libVES.User},
    fieldExtra: {user: true, vaultItems: true},
    fieldSets: [{vaultEntries: {id: true, encData: true, vaultItem: {id: true}}},{type: true, algo: true, publicKey: true}, {unlockableVaultKeys: {id: true, type: true}}],
    getAlgo: function() {
	return this.getField('algo');
    },
    getType: function() {
	return this.getField('type');
    },
    getPublicKey: function() {
	return this.getField('publicKey');
    },
    getPrivateKey: function() {
	return this.getField('privateKey');
    },
    getUnlockedPrivateKey: function() {
	var self = this;
	return this.unlock().then(function(k) {
	    return self.engine().then(function(e) {
		return e.export(k,{opentext:true});
	    });
	});
    },
    getVaultItems: function() {
	return this.getField('vaultItems');
    },
    getExternals: function() {
	return this.getField('externals');
    },
    getUser: function() {
	return this.getField('user');
    },
    resolveVESkey: function(veskey) {
	if (veskey) return Promise.resolve(veskey);
	var self = this;
	return self.getType().then(function(t) {
	    switch (t) {
	    case 'secondary':
	    case 'temp':
	    case 'lost':
		return self.getVaultItems().then(function(vis) {
		    var f = function(vis) {
			if (!vis.length) throw new libVES.Error('InvalidKey','Cannot unlock the secondary key');
			return vis[0].getType().then(function(t) {
			    switch (t) {
				case 'password': return vis[0].get();
				default: return f(vis.slice(1));
			    }
			});
		    };
		    return f(vis);
		});
	    default: throw new libVES.Error('InvalidKey','Cannot unlock the key',{vaultKey: self});
	    }
	});
    },
    unlock: function(veskey) {
	var self = this;
	if (self.wcPriv) return self.wcPriv;
	return self.getId().then(function(id) {
	    if (!self.VES.unlockedKeys[id]) return self.VES.unlockedKeys[id] = self.engine().then(function(m) {
		return self.resolveVESkey(veskey).then(function(v) {
		    return self.getPrivateKey().then(function(prk) {
			return m.import(prk,{password: v});
		    }).catch(function(e) {
			if (e.code != 'Legacy' || !self.VES.unlockLegacyKey) throw e;
			delete(self.VES.unlockedKeys[id]);
			return self.VES.unlockLegacyKey(self,veskey);
		    });
		});
	    });
	    else return self.VES.unlockedKeys[id].catch(function(e) {
		self.VES.unlockedKeys[id] = null;
		return self.unlock(veskey);
	    });
	});
    },
    lock: function() {
	var self = this;
	return this.getId().then(function(id) {
	    delete(self.wcPriv);
	    delete(self.VES.unlockedKeys[id]);
	    return true;
	});
    },
    getPublicCryptoKey: function() {
	if (!this.wcPub) {
	    var self = this;
	    self.wcPub = this.engine().then(function(e) {
		return self.getPublicKey().then(function(pubk) {
		    return e.import(pubk);
		});
	    });
	}
	return this.wcPub;
    },
    engine: function() {
	return this.getAlgo().then(function(algo) {
	    return libVES.getModule(libVES.Algo,algo);
	});
    },
    generate: function(veskey,optns) {
	if (!optns) optns = this.VES.keyOptions;
	var self = this;
	var a;
	var wc = optns && optns.privateKey ? libVES.Algo.acquire(optns.privateKey).then(function(wc) {
	    self.setField('algo',wc.engine.tag);
	    if (!wc.privateKey) throw new libVES.Error('InvalidValue','Private key expected');
	    return wc;
	}) : (optns && (a = libVES.Algo.fromKeyOptions(optns)) ? self.setField('algo', a) : Promise.resolve()).then(function() {
	    return self.engine().then(function(e) {
		return e.generate(optns).then(function(ks) {
		    ks.engine = e;
		    return ks;
		});
	    });
	});
	return Promise.resolve(veskey).then(function(v) {
	    self.wcPub = wc.then(function(ks) {
		return ks.publicKey;
	    });
	    self.setField('publicKey',wc.then(function(ks) {
		return ks.engine.export(ks.publicKey);
	    }));
	    self.wcPriv = wc.then(function(ks) {
		return ks.privateKey;
	    });
	    self.setField('privateKey',wc.then(function(ks) {
		return ks.engine.export(ks.privateKey,{password:v});
	    }));
	    return self;
	});
    },
    encrypt: function(ptxt) {
	var self = this;
	return self.engine().then(function(e) {
	    return self.getPublicCryptoKey().then(function(k) {
		return e.encrypt(k,ptxt).then(function(ctxt) {
		    return libVES.Util.ByteArrayToB64(ctxt);
		});
	    });
	});
    },
    decrypt: function(ctxt) {
	var self = this;
	return self.engine().then(function(e) {
	    return self.unlock().then(function(k) {
		return e.decrypt(k,libVES.Util.B64ToByteArray(ctxt));
	    });
	});
    },
    getVaultEntries: function(details) {
	return this.getField('vaultEntries',{id: true, encData: true, vaultItem: (typeof(details) == 'object' ? details : ((details != null && !details) ? true : {id: true, type: true, meta: true}))});
    },
    rekeyFrom: function(key,veskey) {
	var self = this;
	var old_vis = {};
	return (self.vaultEntries ? self.vaultEntries.then(function(old_ves) {
	    if (old_ves) return old_ves.map(function(ve,i) {
		old_vis[ve.vaultItem.id] = true;
	    });
	}) : Promise.resolve(null)).then(function() {
	    return self.setField('vaultEntries',key.unlock(veskey).then(function() {
		return key.getVaultEntries().then(function(ves) {
		    return Promise.all(ves.map(function(ve) {
			return old_vis[ve.vaultItem.id] ? null : key.decrypt(ve.encData).then(function(ptxt) {
			    return self.encrypt(ptxt).then(function(ctxt) {
				return {
				    vaultItem: {id: ve.vaultItem.id},
				    encData: ctxt
				};
			    });
			}).catch(function(e) {
			    return {
				vaultItem: {id: ve.vaultItem.id},
				"$op": "ignore"
			    };
			});
		    }).filter(function(e, i) {
			return !!e;
		    }));
		});
	    }));
	}).then(function() {
	    return self;
	});
    },
    rekey: function(optns) {
	var self = this;
	return self.getUser().then(function(user) {
	    return self.getExternals().then(function(exts) {
		return (exts && exts.length ? exts[0].toRef().then(function(ext) {
		    ext.user = user;
		    return ext;
		}) : Promise.resolve(user)).then(function(ref) {
		    return self.VES.usersToKeys([ref]);
		});
	    }).then(function(keys) {
		return Promise.all(keys.map(function(key,i) {
		    return key.getVaultEntries().catch(function(e) {
			if (e.code != 'NotFound') throw e;
			key.vaultEntries = undefined;
		    }).then(function() {
			return key.rekeyFrom(self);
		    });
		}));
	    }).then(function(keys) {
		return user.setField('vaultKeys',keys).then(function() {
		    if (!optns) optns = {};
		    optns.refs = {'#/': user};
		    return user.post(null,{vaultEntries: true}, optns).then(function(data) {
			self.setFields(data,false);
			return self.VES.onRekey ? self.VES.onRekey(self).catch(function() {}).then(function() {
			    return self;
			}) : self;
		    });
		});
	    });
	});
    },
    getRecovery: function(myItems) {
	var self = this;
	return self.getType().then(function(t) {
	    switch (t) {
		case 'shadow': case 'recovery':
		    return new libVES.Recovery(self, myItems);
		default: throw new libVES.Error('InvalidValue','Recovery is not applicable for VaultKey type ' + t);
	    }
	});
    },
    getSessionToken: function() {
	var self = this;
	return this.getField('encSessionToken').then(function(tk) {
	    if (!tk) return null;
	    return self.decrypt(tk).then(function(b) {
		return libVES.Util.ByteArrayToString(b);
	    });
	});
    },
    reshareVESkey: function(veskey, optns) {
	var self = this;
	return self.getVaultItems().then(function(vaultItems) {
	    return self.getUser().then(function(user) {
		return Promise.all(vaultItems.map(function(vaultItem,i) {
		    return vaultItem.getType().then(function(t) {
			if (t == 'password') return vaultItem.reshareWith([user], veskey, false).then(function(vi) {
			    return vi.post(undefined, undefined, optns);
			});
		    });
		}));
	    });
	});
    },
    matchVaults: function(vaultKeys) {
	return Promise.resolve(false);
    },
    getKeyOptions: function() {
	var self = this;
	return self.engine().then(function(e) {
	    return self.getPublicCryptoKey().then(function(pub) {
		return e.getKeyOptions(pub);
	    });
	});
    }
});

libVES.VaultItem.prototype = new libVES.Object({
    apiUri: 'vaultItems',
    fieldList: {id: true, deleted: true, file: true},
    fieldClass: {vaultKey: libVES.VaultKey, file: libVES.File},
    fieldSets: [{type: true, meta: true},{vaultEntries: {id: true, encData: true, vaultKey: {id: true, type: true, user: {id: true}, algo: true, externals: {id: true}}}},{vaultKey: true, file: true}],
    defaultCipher: 'AES256GCM',
    getRaw: function() {
	var self = this;
	var f = function(vaultEntries, vaultKeys) {
	    var i = 0;
	    var fn = function() {
		if (vaultEntries) for (; i < vaultEntries.length; i++) {
		    var k, d;
		    if ((d = vaultEntries[i].encData) != null && vaultKeys[(k = vaultEntries[i].vaultKey).id]) {
			i++;
			return new libVES.VaultKey(k, self.VES).decrypt(d).catch(fn);
		    }
		}
		return Promise.reject(new libVES.Error('InvalidKey',"No unlocked key to decrypt the item",{vaultItem: self}));
	    };
	    return fn();
	};
	var vaultEntries = [];
	if (self.vaultEntryByKey) for (var k in self.vaultEntryByKey) vaultEntries.push(self.vaultEntryByKey[k]);
	return f(vaultEntries, self.VES.unlockedKeys).catch(function() {
	    return self.getVaultEntries().then(function(vaultEntries) {
		return f(vaultEntries, self.VES.unlockedKeys).catch(function() {
		    return self.VES.getUnlockableKeys().then(function(vks) {
			return f(vaultEntries, vks);
		    });
		});
	    });
	});
    },
    get: function() {
	var self = this;
	return this.getRaw().then(function(buf) {
	    return self.parse(buf);
	});
    },
    getType: function() {
	return this.getField('type');
    },
    getMeta: function() {
	return this.getField('meta');
    },
    getVaultEntries: function() {
	var self = this;
	return this.getField('vaultEntries').then(function(ves) {
	    if (ves) for (var i = 0; i < ves.length; i++) self.vaultEntryByKey[ves[i].vaultKey.id] = ves[i];
	    return ves;
	});
    },
    getVaultKey: function() {
	return this.getField('vaultKey');
    },
    getFile: function() {
	return this.getField('file');
    },
    getDeleted: function() {
	return this.getField('deleted');
    },
    parse: function(buf) {
	var self = this;
	return this.getType().then(function(type) {
	    return libVES.getModule(libVES.VaultItem.Type,type).then(function(m) {
		return m.parse.call(self,buf);
	    }).catch(function(e) {
		return new Uint8Array(buf);
	    });
	});
    },
    build: function(data) {
	var self = this;
	return this.getType().then(function(type) {
	    return libVES.getModule(libVES.VaultItem.Type,type).then(function(m) {
		return m.build.call(self,data);
	    });
	});
    },
    resolveRaw: function(val) {
	return (val == null ? this.getRaw() : this.build(val));
    },
    shareWith: function(usrs,val,save) {
	var self = this;
	return self.VES.usersToKeys(usrs).then(function(ks) {
	    return (val == null || self.id ? self.getVaultEntries().then(function(ves) {
		var k_ves = {};
		var k_used = {};
		for (var j = 0; j < ves.length; j++) k_ves[ves[j].vaultKey.id] = ves[j];
		return Promise.all(ks.map(function(k,j) {
		    return k.getId().then(function(k_id) {
			k_used[k_id] = true;
			return k_ves[k_id];
		    }).catch(function(){});
		})).then(function(old_ves) {
		    for (var k_id in k_ves) if (!k_used[k_id]) old_ves.push(k_ves[k_id]);
		    return old_ves;
		});
	    }) : Promise.resolve([])).then(function(old_ves) {
		var new_ves = [];
		var set_ves = [];
		var valr = null;
		return Promise.all(ks.map(function(k,j) {
		    return new_ves[j] = (old_ves[j] || (valr != null ? valr : valr = self.resolveRaw(val)).then(function(v) {
			return k.encrypt(v).then(function(ctext) {
			    return (function(refs) {
				if (refs) for (var i in refs) if (refs[i] === k) return Promise.resolve({'$ref':i});
				return k.postData(null,refs);
			    })(libVES.Object._refs).then(function(pd) {
				return set_ves.push({vaultKey: pd, encData: ctext});
			    });
			});
		    }));
		})).then(function() {
		    return Promise.all(old_ves.slice(ks.length).map(function(ve,j) {
			return (new libVES.VaultKey(ve.vaultKey,self.VES)).matchVaults(ks).then(function(f) {
			    if (f === false) set_ves.push({vaultKey: {id: ve.vaultKey.id}, '$op': 'delete'});
			});
		    }));
		}).then(function() {
		    if (!set_ves.length) return save = false;
		    return self.setField('vaultEntries',set_ves);
		});
	    });
	}).then(function() {
	    if (save || save === undefined) return self.post(undefined, undefined, ((save instanceof Object) ? save : undefined)).then(function() {
		delete(self.vaultEntries);
		return self;
	    });
	    return self;
	});
    },
    reshareWith: function(share,val,save) {
	var self = this;
	return self.VES.usersToKeys(share).then(function(new_ks) {
	    return self.getShareVaultKeys().then(function(curr_ks) {
		return Promise.all(curr_ks.map(function(k,i) {
		    return k.getId();
		})).then(function(curr_ids) {
		    var m_curr_ks = {};
		    for (var i = 0; i < curr_ks.length; i++) m_curr_ks[curr_ids[i]] = curr_ks[i];
		    return Promise.all(new_ks.map(function(k,i) {
			return k.getId().catch(function(e) {
			    if (e.code != 'NotFound') throw e;
			});
		    })).then(function(new_ids) {
			for (var i = 0; i < new_ks.length; i++) if (!m_curr_ks[new_ids[i]]) curr_ks.push(m_curr_ks[new_ids[i]] = new_ks[i]);
			return self.shareWith(curr_ks,val,save);
		    });
		});
	    });
	});
    },
    getShareVaultKeys: function() {
	var self = this;
	return this.getVaultEntries().then(function(vaultEntries) {
	    return vaultEntries.map(function(e,i) {
		return new libVES.VaultKey(e.vaultKey,self.VES);
	    });
	});
    },
    getShareList: function() {
	var self = this;
	return this.getShareVaultKeys().then(function(vaultKeys) {
	    var uids = {};
	    return Promise.all(vaultKeys.map(function(e,i) {
		return e.getExternals().then(function(exts) {
		    if (exts && exts.length) return exts[0];
		    return e.getUser().then(function(u) {
			return u.getId().then(function(uid) {
			    if (uids[uid]) return null;
			    uids[uid] = true;
			    return u;
			});
		    });
		});
	    })).then(function(lst) {
		var rs = [];
		for (var i = 0; i < lst.length; i++) if (lst[i]) rs.push(lst[i]);
		return rs;
	    });
	});
    },
    delete: function() {
	var self = this;
	return self.getId().then(function(id) {
	    return this.VES.post(self.apiUri + '/' + id, {'$op': 'delete'});
	});
    }
});
libVES.VaultItem.Type = {
    _detect: function(data) {
	if (typeof(data) == 'object') {
	    if (data instanceof libVES.Cipher) return 'file';
	    throw new libVES.Error('Internal','Unknown vault item data type');
	} else return 'string';
    },
    string: {
	parse: function(buf) {
	    return libVES.Util.ByteArrayToString(buf);
	},
	build: function(data) {
	    return libVES.Util.StringToByteArray(String(data));
	}
    },
    file: {
	parse: function(buf) {
	    var self = this;
	    return this.getMeta().then(function(meta) {
		var ci = libVES.Cipher[meta.a || self.defaultCipher];
		return new ci(new Uint8Array(buf));
	    });
	},
 	build: function(data) {
	    if (!(data instanceof libVES.Cipher)) throw new libVES.Error('InvalidData',"Content of a VaultItem type 'file' must be libVES.Cipher");
	    this.setField('meta', data.getMeta());
	    return data.getSecret();
	}
    },
    secret: {
	parse: function(buf) {
	    var self = this;
	    return this.getMeta().then(function(meta) {
		return {value: buf, meta: meta};
	    });
	},
	build: function(data) {
	    this.setField('meta',data.meta);
	    return data.value;
	}
    }
};
libVES.VaultItem.Type.password = libVES.VaultItem.Type.string;

libVES.File.prototype = new libVES.Object({
    apiUri: 'files',
    fieldList: {id: true, externals: true},
    fieldClass: {externals: libVES.External, creator: libVES.User},
    getExternals: function() {
	return this.getField('externals');
    },
    getCreator: function() {
	return this.getField('creator');
    },
});

libVES.External.prototype = new libVES.Object({
    apiUri: 'externals',
    fieldList: {id: true},
    getDomain: function() {
	return this.getField('domain');
    },
    getExternalId: function() {
	return this.getField('externalId');
    },
    toRef: function() {
	return Promise.all([this.getDomain(),this.getExternalId()]).then(function(r) {
	    return {domain: r[0], externalId: r[1]};
	});
    }
});


libVES.Event.prototype = new libVES.Object({
    apiUri: 'events',
    fieldList: {id: true, recordedAt: true, vaultKey: true, vaultItem: true, user: true},
    fieldClass: {vaultKey: libVES.VaultKey, vaultItem: libVES.VaultItem, user: libVES.User, creator: libVES.User, session: libVES.Session},
    getType: function() {
	return this.getField('type');
    },
    getVaultKey: function() {
	return this.getField('vaultKey');
    },
    getVaultItem: function() {
	return this.getField('vaultItem');
    },
    getUser: function() {
	return this.getField('user');
    },
    getCreator: function() {
	return this.getField('creator');
    },
    getSession: function() {
	return this.getField('session');
    }
});

libVES.Session.prototype = new libVES.Object({
    apiUri: 'sessions',
    fieldList: {id: true, createdAt: true, expiresAt: true, vaultKey: true, user: true},
    fieldClass: {vaultKey: libVES.VaultKey, user: libVES.User},
    getVaultKey: function() {
	return this.getField('vaultKey');
    },
    getUser: function() {
	return this.getField('user');
    },
    getRemote: function() {
	return this.getField('remote');
    },
    getUserAgent: function() {
	return this.getField('userAgent');
    }
});

libVES.Domain.prototype = new libVES.Object({
    apiUri: 'domains',
    fieldList: {id: true},
    fieldClass: {vaultKeys: libVES.VaultKey, vaultItems: libVES.VaultItem, creator: libVES.User},
    fieldSets: [{vaultItems: {id: true, type: true, file: {creator: true, externals: true}, deleted: true}, vaultKeys: {id: true, type: true, user: {id: true}, algo: true, externals: {id: true}}}],
    getCreator: function() {
	return this.getField('creator');
    },
    getVaultKeys: function() {
	return this.getField('vaultKeys');
    },
    getVaultItems: function() {
	return this.getField('vaultItems');
    }
});


libVES.Algo = {
    ECDH: {
	tag: 'ECDH',
	defaultCurve: 'P-521',
	decrypt: function(k,buf) {
	    var b = new Uint8Array(buf);
	    var data = libVES.Util.ASN1.decode(b);
	    if (!data || data.length < 2) throw new libVES.Error('InvalidValue','Error parsing ECIES: Invalid ciphertext?');
	    return libVES.Util.ASN1.import(b.slice(0,b.byteLength - data[1].byteLength),{decoded:data[0]}).then(function(eKey) {
		return libVES.Algo.ECDH.cipher(k,eKey).then(function(ci) {
		    return ci.decrypt(data[1],true).catch(function(e) {
			if (e instanceof libVES.Error) throw e;
			throw new libVES.Error('InvalidValue','Error decrypting ECIES payload: Invalid ciphertext?',{error:e});
		    });
		}).catch(function(e) {
		    if (e instanceof libVES.Error) throw e;
		    throw new libVES.Error('InvalidValue','Error negotiating DH cipher key: Invalid ciphertext?',{error:e});
		});
	    }).catch(function(e) {
		if (e instanceof libVES.Error) throw e;
		throw new libVES.Error('InvalidValue','Error loading EC ephemeral: Invalid ciphertext?',{error:e});
	    });
	},
	encrypt: function(k,buf) {
	    return libVES.Algo.ECDH.eKey(k).then(function(eKey) {
		return libVES.Algo.ECDH.cipher(eKey.privateKey,k).then(function(ci) {
		    return Promise.all([
			((k instanceof CryptoKey)
			    ? crypto.subtle.exportKey('spki',eKey.publicKey)
			    : libVES.Algo.ECDH.wasm().then(function(wasm) {
				return libVES.Util.Key.toPKCS(wasm.getpub(eKey.publicKey), null, libVES.Algo.ECDH.asn1hdr(wasm.getoid(eKey.publicKey)));
			    })
			),
			ci.encrypt(buf,true)
		    ]).then(function(bufs) {
			var buf = new Uint8Array(bufs[0].byteLength + bufs[1].byteLength);
			buf.set(new Uint8Array(bufs[0]),0);
			buf.set(new Uint8Array(bufs[1]),bufs[0].byteLength);
			return buf;
		    });
		}).catch(function(e) {
		    if (e instanceof libVES.Error) throw e;
		    throw new libVES.Error('InvalidValue','Error generating ECIES DH cipher',{error:e});
		});
	    });
	},
	eKey: function(pub) {
	    if (pub instanceof CryptoKey) return crypto.subtle.exportKey('jwk',pub).then(function(kdata) {
		return crypto.subtle.generateKey({name:'ECDH',namedCurve: kdata.crv},true,['deriveKey','deriveBits']).catch(function(e) {
		    throw new libVES.Error('InvalidValue','Error generating ECIES ephemeral',{error: e});
		});
	    });
	    return this.wasm().then(function(wasm) {
		var k = wasm.init(pub.curve);
		wasm.generate(k);
		return wasm.privpub(k);
	    });
	},
	derive: function(priv, pub) {
	    if (pub instanceof CryptoKey) return Promise.resolve(libVES.Algo.ECDH.curveBytes[pub.algorithm.namedCurve] || crypto.subtle.exportKey('jwk', pub).then(function(jwk) {
		return jwk.x.length * 3 / 4;
	    })).then(function(len) {
		return crypto.subtle.deriveBits({name: 'ECDH', public: pub}, priv, 8 * len);
	    });
	    return this.wasm().then(function(wasm) {
		return wasm.derive(priv, pub);
	    });
	},
	cipher: function(priv, pub) {
	    return libVES.getModule(libVES.Cipher,'AES256GCMp').then(function(m) {
		return libVES.Algo.ECDH.derive(priv, pub).then(function(raw) {
		    return crypto.subtle.digest({name:'SHA-384'},raw).then(function(buf) {
			return new m(new Uint8Array(buf).slice(0,m.prototype.keySize + m.prototype.ivSize));
		    });
		});
	    });
	},
	curveBytes: {
	    'P-256': 32,
	    'P-384': 48,
	    'P-521': 66
	},
	import: function(data,optns) {
	    return libVES.Util.PEM.import(data,optns);
	},
	export: function(data,optns) {
	    if (data instanceof CryptoKey) switch (data.type) {
		case 'private': return libVES.Util.PKCS8.encode(data, optns);
		case 'public':
		    return crypto.subtle.exportKey('spki', data).then(function(der) {
			var asn = libVES.Util.ASN1.decode(new Uint8Array(der))[0];
			switch (String(asn[0][0])) {
			    case '1.2.840.10045.2.1':
				return der;
			    case '1.3.132.1.12':
			    case '1.3.132.112': // Firefox on Mac - apparently a misspelling of the former
				asn[0][0] = new libVES.Util.OID('1.2.840.10045.2.1');
				asn[1].ASN1type = 3;
				return libVES.Util.ASN1.encode([asn]);
			    default:
				console.log(asn);
				throw new libVES.Error('Internal', 'Unexpected EC pubkey format');
			}
			return der;
		    }).then(function(der) {
			return libVES.Util.PEM.encode(der, 'PUBLIC KEY');
		    });
	    } else return this.wasm().then(function(wasm) {
		return libVES.Util.Key.export(wasm.getpub(data), wasm.getpriv(data), libVES.Algo.ECDH.asn1hdr(wasm.getoid(data)), optns);
	    });
	    throw new libVES.Error('Internal', "Unknown type of key object");
	},
	generate: function(optns) {
	    var op = {name: 'ECDH', namedCurve: this.defaultCurve};
	    if (optns) for (var k in optns) op[k] = optns[k];
	    return crypto.subtle.generateKey(op, true, ['deriveKey', 'deriveBits']).catch(function(e) {
		console.log('ECDH generateKey failed, trying wasm...', e, op);
		return libVES.Algo.ECDH.wasm().then(function(wasm) {
		    var k = wasm.init(op.namedCurve);
		    if (!k || !wasm.generate(k)) throw new libVES.Error('InvalidValue', 'WasmECDH key generation error (unsupported curve?)');
		    return wasm.privpub(k);
		});
	    });
	},
	getPublic: function(priv) {
	    return crypto.subtle.exportKey('jwk',priv).then(function(k) {
		return crypto.subtle.importKey('jwk',{
		    crv: k.crv,
		    ext: true,
		    key_ops: [],
		    kty: 'EC',
		    x: k.x,
		    y: k.y
		},{name:'ECDH'},true,[]);
	    });
	},
	getKeyOptions: function(key) {
	    if (key instanceof CryptoKey) return crypto.subtle.exportKey('jwk', key).then(function(k) {
		return {namedCurve: k.crv};
	    });
	    return {namedCurve: k.curve};
	},
	asn1hdr: function(oid) {
	    return [new libVES.Util.OID(this.OID), new libVES.Util.OID(oid)];
	},
	wasm: function() {
	    if (!this.wasmP) this.wasmP = (typeof(WasmECDH) == 'function' ? Promise.resolve()
		: libVES.Util.loadWasm(WasmECDHinit.baseUrl + 'WasmECDH.js')
	    ).then(function() {
		return WasmECDH(WasmECDHinit);
	    });
	    return this.wasmP;
	},
	importWasm: function(curve, der) {
	    return libVES.Algo.ECDH.wasm().then(function(wasm) {
		return libVES.Util.Key.fromDER(der, function(pub, priv, pubBits) {
		    var k = wasm.init(curve);
		    if (!k) return null;
		    if (pub) wasm.setpub(k, pub);
		    if (priv) wasm.setpriv(k, priv);
		    return k;
		});
	    });
	},
	getMethods: function() {
	    var crvs = ['P-256', 'P-384', 'P-521'];
	    return this.wasm().then(function(wasm) {
		return crvs.concat(wasm.getcurves());
	    }).catch(function(e) {
		return crvs;
	    }).then(function(crvs) {
		return crvs.map(function(algo) {
		    return {namedCurve: algo};
		});
	    });
	},
	OID: '1.2.840.10045.2.1'
    },
    RSA: {
	tag: 'RSA',
	decrypt: function(k,buf) {
	    var self = libVES.Algo.RSA;
	    var maxl = (k.algorithm.modulusLength + 7) >> 3;
	    if (buf.byteLength <= maxl) return self.decryptRSA(k,buf);
	    var b = new Uint8Array(buf);
	    return self.decryptRSA(k,b.slice(0,maxl)).then(function(ck) {
		return self.cipher(ck).then(function(ci) {
		    return ci.decrypt(b.slice(maxl),true);
		});
	    });
	},
	decryptRSA: function(k,buf) {
	    return crypto.subtle.decrypt('RSA-OAEP',k,buf).catch(function(e) {
		throw new libVES.Error('InvalidValue','Error decrypting RSA OAEP: Invalid ciphertext?',{error:e});
	    });
	},
	encrypt: function(k,buf) {
	    var self = libVES.Algo.RSA;
	    var maxl = ((k.algorithm.modulusLength + 7) >> 3) - 48;
	    if (buf.byteLength <= maxl) return self.encryptRSA(k,buf);
	    return self.cipher().then(function(ci) {
		return Promise.all([
		    ci.getSecret().then(function(sk) {
			return self.encryptRSA(k,sk);
		    }),
		    ci.encrypt(buf,true)
		]).then(function(bufs) {
		    var rs = new Uint8Array(bufs[0].byteLength + bufs[1].byteLength);
		    rs.set(new Uint8Array(bufs[0]),0);
		    rs.set(new Uint8Array(bufs[1]),bufs[0].byteLength);
		    return rs;
		});
	    });
	},
	encryptRSA: function(k,buf) {
	    return crypto.subtle.encrypt('RSA-OAEP',k,buf).catch(function(e) {
		throw new libVES.Error('InvalidValue','Error encrypting RSA OAEP: Buffer is too long?',{error:e});
	    });;
	},
	cipher: function(secret) {
	    return libVES.getModule(libVES.Cipher,'AES256GCMp').then(function(cls) {
		return new cls(secret);
	    });
	},
	import: function(data,optns) {
	    return libVES.Util.PEM.import(data,optns);
	},
	export: function(data,optns) {
	    if (data instanceof CryptoKey) switch (data.type) {
		case 'private': return libVES.Util.PKCS8.encode(data,optns);
		case 'public': return libVES.Util.PKCS1.encode(data,optns);
	    }
	    throw new libVES.Error('Internal',"Unknown type of key object");
	},
	generate: function(optns) {
	    var op = {name:'RSA-OAEP', modulusLength:4096, publicExponent:new Uint8Array([1,0,1]), hash:'SHA-256'};
	    if (optns) for (var k in optns) op[k] = optns[k];
	    return crypto.subtle.generateKey(op,true,['decrypt','encrypt']);
	},
	getPublic: function(priv) {
	    return crypto.subtle.exportKey('jwk',priv).then(function(k) {
		return crypto.subtle.importKey('jwk',{
		    alg: k.alg,
		    e: k.e,
		    ext: true,
		    key_ops: ['encrypt'],
		    kty: 'RSA',
		    n: k.n
		},{name:'RSA-OAEP',hash:'SHA-256'},true,['encrypt']);
	    });
	},
	getKeyOptions: function(key) {
	    return crypto.subtle.exportKey('jwk', key).then(function(k) {
		return {modulusLength: ((k.n.length * 6) >> 3) * 8, publicExponent: libVES.Util.B64ToByteArray(k.e)};
	    });
	},
	getMethods: function() {
	    var bits = [];
	    var s = 128;
	    for (var i = 640; i <= 16384; i += s) {
		bits.push(i);
		if (i / s >= 8) s <<= 1;
	    }
	    return Promise.resolve(bits.map(function(b) {
		    return {modulusLength: b};
	    }));
	},
    },
    RSA_PKCS1_15: {
	tag: 'RSA_PKCS1_15',
	import: function(data,optns) {
	    throw new libVES.Error('Legacy','RSA with PKCS#1 1.5 padding is not supported');
	}
    },
    acquire: function(key,optns) {
	return Promise.resolve(key).then(function(k) {
	    if (typeof(CryptoKey) != 'undefined' && (k instanceof CryptoKey)) return k;
	    else if (typeof(k) == 'object') {
		if (k.privateKey) return k.privateKey;
		else if (k.publicKey) return k.publicKey;
		else throw new libVES.Error('Internal','Unknown key format');
	    } else return libVES.Util.PEM.import(k,optns);
	}).then(function(k) {
	    switch (k.algorithm.name) {
		case 'RSA-OAEP': return libVES.getModule(libVES.Algo,'RSA').then(function(e) {
		    var rs = {engine: e};
		    switch (k.type) {
			case 'private':
			    rs.privateKey = k;
			    return e.getPublic(k).then(function(pubk) {
				rs.publicKey = pubk;
				return rs;
			    });
			case 'public':
			    rs.publicKey = k;
			    return rs;
			default: throw new libVES.Error('Internal','Unsupported key type: ' + k.type);
		    }
		});
		default: throw new libVES.Error('Internal','Unsupported key algorithm: ' + k.algorithm.name);
	    }
	});
    },
    fromKeyOptions: function(optns) {
	if (optns.namedCurve) return 'ECDH';
	if (optns.oqsAlgo) return 'OQS';
	if (optns.modulusLength) return 'RSA';
    },
    toString: function() {
	return 'libVES.Algo';
    },
    all: ['RSA', 'ECDH']
};

WasmECDHinit = {
    Key: function() {},
    init: function(curve) {
	var a = new Uint8Array(libVES.Util.StringToByteArray(curve));
	var arg1 = this.buf(a.byteLength + 1);
	arg1.set(a);
	arg1.set([0], a.byteLength);
	var ptr = this._WasmECDH_new(arg1.byteOffset);
	if (!ptr) return null;
	var key = new this.Key();
	key.ptr = ptr;
	key.curve = this.getcurve(key);
	return key;
    },
    buf: function(len) {
	return new Uint8Array(this.asm.memory.buffer, this._WasmECDH_buf, len);
    },
    str: function(ptr) {
	if (!ptr) return null;
	var b = new Uint8Array(this.asm.memory.buffer, ptr);
	return libVES.Util.ByteArrayToString(b.slice(0, b.indexOf(0)));
    },
    generate: function(key) {
	key.private = true;
	return this._WasmECDH_generate(key.ptr);
    },
    setpub: function(key, pub) {
	var arg1 = this.buf(pub.byteLength);
	arg1.set(pub);
	return this._WasmECDH_setpub(key.ptr, arg1.byteOffset, arg1.byteLength);
    },
    setpriv: function(key, priv) {
	key.private = true;
	var arg1 = this.buf(priv.byteLength);
	arg1.set(priv);
	return this._WasmECDH_setpriv(key.ptr, arg1.byteOffset, arg1.byteLength);
    },
    getpub: function(key) {
	var l = this._WasmECDH_getpub(key.ptr);
	if (l <= 0) return null;
	var rs = new Uint8Array(l);
	rs.set(this.buf(l));
	return rs;
    },
    getpriv: function(key) {
	if (!key.private) return null;
	var l = this._WasmECDH_getpriv(key.ptr);
	if (l <= 0) return null;
	var rs = new Uint8Array(l);
	rs.set(this.buf(l));
	return rs;
    },
    getcurve: function(key) {
	return this.str(this._WasmECDH_getcurve(key.ptr));
    },
    getoid: function(key) {
	return this.str(this._WasmECDH_getoid(key.ptr));
    },
    derive: function(priv, pub) {
	var l = this._WasmECDH_derive(priv.ptr, pub.ptr);
	if (l <= 0) return null;
	var rs = new Uint8Array(l);
	rs.set(this.buf(l));
	return rs;
    },
    privpub: function(priv) {
	var pub = new this.Key();
	pub.ptr = priv.ptr;
	pub.curve = priv.curve;
	return {privateKey: priv, publicKey: pub};
    },
    free: function(key) {
	this._WasmECDH_free(key.ptr);
    },
    getcurves: function() {
	var lst = this._WasmECDH_listinit(1024);
	var crvs = [];
	var c;
	for (var i = 0; (c = this._WasmECDH_listget(lst, i)); i++) {
	    var crv = this.str(c);
	    if (crv == '') continue;
	    crvs.push(crv);
	}
	this._WasmECDH_listfree(lst);
	return crvs;
    },
    locateFile: function(file) {
       return this.baseUrl + file;
    },
    baseUrl: 'https://ves.host/pub/'
};

libVES.Cipher = function(data) {
    for (var k in data) this[k] = data[k];
};
libVES.Cipher.prototype = {
    init: function(secret) {
	if (!secret) {
	    secret = new Uint8Array(this.keySize + this.ivSize);
	    crypto.getRandomValues(secret);
	} else if (secret.length < this.keySize) throw new libVES.Error('Internal','Invalid cipher key data');
	this.key = this.buildKey(secret.slice(0,this.keySize));
	this.IV = Promise.resolve(secret.slice(0,this.keySize + this.ivSize).slice(-this.ivSize));
	try {
	    this.meta = JSON.parse(libVES.Util.ByteArrayToString(secret.slice(this.keySize + this.ivSize)));
	} catch(e) {
	    this.meta = {};
	}
    },
    getSecret: function() {
	var self = this;
	var meta = null;
	if (this.meta) for (var k in this.meta) {
	    meta = libVES.Util.StringToByteArray(JSON.stringify(this.meta));
	    break;
	}
	var buf = new Uint8Array(this.keySize + this.ivSize + (meta ? meta.byteLength : 0));
	return Promise.all([(this.algo ? this.key.then(function(key) {
	    return crypto.subtle.exportKey("raw", key);
	}) : this.key), this.IV]).then(function(data) {
	    buf.set(new Uint8Array(data[0]), 0);
	    buf.set(new Uint8Array(data[1]), data[0].byteLength);
	    if (meta) buf.set(new Uint8Array(meta), data[0].byteLength + data[1].byteLength);
	    return buf;
	});
    },
    getMeta: function() {
	return {a: this.a};
    },
    buildKey: function(key) {
	if (!this.algo) return Promise.resolve(key);
	return crypto.subtle.importKey('raw',key,this.algo,true,['encrypt','decrypt']);
    },
    process: function(buf,final,callbk,sizefn) {
	var self = this;
	buf = new Uint8Array(buf);
	if (this.processBuf) {
	    var b = new Uint8Array(buf.byteLength + this.processBuf.byteLength);
	    b.set(this.processBuf,0);
	    b.set(buf,this.processBuf.byteLength);
	    buf = b;
	}
	var over = !final;
	var fn = function(cprev) {
	    var chunkSize = sizefn();
	    if ((!buf.byteLength && over) || (!final && (!chunkSize || buf.byteLength < chunkSize))) {
		self.processBuf = buf.byteLength ? buf : null;
		return Promise.resolve(cprev);
	    }
	    var cl = (chunkSize && chunkSize < buf.byteLength) ? chunkSize : buf.byteLength;
	    if (!chunkSize || cl < chunkSize) over = true;
	    return callbk(buf.slice(0, cl)).then(function(ctext) {
		buf = buf.slice(cl);
		var r = new Uint8Array(cprev.byteLength + ctext.byteLength);
		r.set(cprev, 0);
		r.set(new Uint8Array(ctext), cprev.byteLength);
		return r;
	    }).then(fn);
	};
	return fn(new Uint8Array(0));
    },
    encryptChunk: function(buf) {
	return Promise.all([this.key,this.algoInfo()]).then(function(info) {
	    return crypto.subtle.encrypt(info[1],info[0],buf);
	});
    },
    decryptChunk: function(buf) {
	return Promise.all([this.key,this.algoInfo()]).then(function(info) {
	    return crypto.subtle.decrypt(info[1],info[0],buf);
	});
    },
    algoInfo: function() {
	return Promise.resolve(this.algo);
    },
    encrypt: function(buf,final) {
	return this.process(buf,final,this.encryptChunk.bind(this),(function() { return this.chunkSizeP; }).bind(this));
    },
    decrypt: function(buf,final) {
	return this.process(buf,final,this.decryptChunk.bind(this),(function() { return this.chunkSizeC; }).bind(this));
    }
};

libVES.Cipher.NULL = function(rec) {
    this.init(rec);
}

libVES.Cipher.AES = function(data) {
    for (var k in data) this[k] = data[k];
};

libVES.Cipher.AES256CBC = function(rec) {
    this.init(rec);
};

libVES.Cipher.AES256CFB = function(rec) {
    this.init(rec);
};

libVES.Cipher.AES256GCM = function(rec) {
    this.init(rec);
};

libVES.Cipher.AES256GCMp = function(rec) {
    this.init(rec);
};

libVES.Cipher.AES256GCM1K = function(rec) {
    this.init(rec);
};

libVES.Cipher.NULL.prototype = new libVES.Cipher({
    a: 'NULL',
    keySize: 0,
    ivSize: 0
});

libVES.Cipher.AES.prototype = new libVES.Cipher({
    keySize: 32,
    ivSize: 32,
    algoInfo: function() {
	var self = this;
	return this.IV.then(function(iv) {
	    return {name: self.algo, iv: iv};
	});
    }
});

libVES.Cipher.AES256CBC.prototype = new libVES.Cipher.AES({
    a: 'AES256CBC',
    algo: 'AES-CBC',
    keySize: 32,
    ivSize: 16
});

libVES.Cipher.AES256CBC.import = function(args,chain,optns) {
    return chain('import').then(function(buf) {
	return crypto.subtle.decrypt({name: 'AES-CBC', iv: args},optns.key,buf);
    });
};
libVES.Cipher.AES256CBC.export = function(chain,optns) {
    var args = new Uint8Array(16);
    crypto.getRandomValues(args);
    return crypto.subtle.encrypt({name: 'AES-CBC', iv: args},optns.key,optns.content).then(function(buf) {
	return chain('export',{content: buf}).then(function() {
	    return [new libVES.Util.OID('2.16.840.1.101.3.4.1.42'), args];
	});
    });
};
libVES.Cipher.AES256CBC.info = function(chain,optns) {
    return Promise.resolve({algorithm: {name: 'AES-CBC', length: 256}});
};

libVES.Cipher.AES256CFB.prototype = new libVES.Cipher.AES({
    a: 'AES256CFB',
    algo: 'AES-CFB'
});

libVES.Cipher.AES256GCM.prototype = new libVES.Cipher.AES({
    algo: 'AES-GCM',
    keySize: 32,
    ivSize: 12
});

libVES.Cipher.AES256GCMp.prototype = new libVES.Cipher.AES({
    a: 'AES256GCMp',
    algo: 'AES-GCM',
    keySize: 32,
    ivSize: 12,
    padSize: 32,
    encryptChunk: function(buf0) {
	var buf = new Uint8Array(buf0);
	var pad = this.padSize - (buf.byteLength % this.padSize) - 1;
	var bufp = new Uint8Array(buf.byteLength + pad + 1);
	bufp[0] = pad;
	bufp.set(buf,1);
	return libVES.Cipher.prototype.encryptChunk.call(this,bufp);
    },
    decryptChunk: function(buf) {
	return libVES.Cipher.prototype.decryptChunk.call(this,buf).then(function(bufp0) {
	    var bufp = new Uint8Array(bufp0);
	    return bufp.slice(1,bufp.byteLength - bufp[0]);
	});
    }
});

libVES.Cipher.AES256GCM1K.prototype = new libVES.Cipher.AES({
    a: 'AES256GCM1K',
    algo: 'AES-GCM',
    keySize: 32,
    ivSize: 12,
    chunkSizeP: 1024,
    chunkSizeC: 16,
    algoInfo: function() {
	var self = this;
	if (!this.lastGCM) throw new libVES.Error('Internal', 'GCM HMAC is not buffered');
	return this.IV.then(function(iv) {
	    var buf = new Uint8Array(iv.byteLength + self.lastGCM.byteLength);
	    buf.set(new Uint8Array(iv), 0);
	    buf.set(new Uint8Array(self.lastGCM), iv.byteLength);
	    return libVES.Util.Hash.SHA256.hash(buf).then(function(sha) {
		return { name: self.algo, iv: sha.slice(0, self.ivSize) };
	    });
	});
    },
    encryptChunk: function(buf) {
	var self = this;
	var seed = null;
	if (!this.lastGCM) {
	    seed = new Uint8Array(16);
	    crypto.getRandomValues(seed);
	    this.lastGCM = seed;
	}
	return Promise.all([this.key, this.algoInfo()]).then(function(info) {
	    return crypto.subtle.encrypt(info[1], info[0], buf).then(function(ctext) {
		self.lastGCM = new Uint8Array(ctext.slice(-16));
		if (!seed) return ctext;
		var rs = new Uint8Array(seed.byteLength + ctext.byteLength);
		rs.set(seed, 0);
		rs.set(new Uint8Array(ctext), seed.byteLength);
		return rs;
	    });
	});
    },
    decryptChunk: function(buf) {
	var self = this;
	if (buf.byteLength < 16) throw new libVES.Error('Crypto', 'Invalid ciphertext block size');
	var rs;
	if (this.lastGCM) {
	    rs = Promise.all([this.key, this.algoInfo()]).then(function(info) {
		self.lastGCM = buf.slice(-16);
		return crypto.subtle.decrypt(info[1], info[0], buf);
	    });
	} else {
	    rs = Promise.resolve(new Uint8Array(0));
	    this.chunkSizeC = 1040;
	    this.lastGCM = buf;
	}
	return rs;
    }
});

libVES.Scramble = {
    RDX: function(x) {
	this.size = x;
    }
};
libVES.Scramble.RDX.prototype = {
    tag: 'RDX1.2',
    name: 'RDX 1.2 Shamir',
    getBases: function(n) {
	var rs = [];
	for (var b = 0; rs.length < n; b++) if (b % 4) rs.push(b);
	return rs;
    },
    getCv: function(b) {
	var rs = [libVES.Math.pad(1)];
	b = Number(b);
	for (var i = 1; i < this.size; i++) rs.push(libVES.Math.mul(rs[rs.length - 1], b));
	return rs;
    },
    toVector: function(sc) {
	var u = [sc];
	for (var i = 1; i < this.size; i++) {
	    u[i] = new Uint8Array(sc.length);
	    crypto.getRandomValues(u[i]);
	}
	var v = [];
	for (var i = 0; i < this.size - 1; i++) v[i] = (function(ui,ui1) {
	    var kbuf = new Uint8Array(32);
	    kbuf.set(ui1);
	    return crypto.subtle.importKey('raw',kbuf,'AES-CTR',true,['encrypt','decrypt']).then(function(k) {
		return crypto.subtle.encrypt({name: 'AES-CTR', counter: new Uint8Array(16).fill(0), length: 128},k,ui).then(function(ctx) {
		    var ctxt = new Uint8Array(ctx);
		    var rs = new Uint8Array(ctxt.length + 1);
		    rs.set(ctxt);
		    rs[ctxt.length] = 1;
		    return rs;
		});
	    });
	})(u[i],u[i + 1]);
	var vi = new Uint8Array(u[i].length + 1);
	vi.set(u[i]);
	vi[u[i].length] = 1;
	v[i] = Promise.resolve(vi);
	return Promise.all(v);
    },
    fromVector: function(vec) {
	var v = [];
	var rs = new Promise(function(resolve,reject) {
	    for (var i = 0; i < vec.length; i++) {
		if (vec[i][vec[i].length - 1] != 1) return reject(new libVES.Error('Internal','Invalid recovery vector'));
		v[i] = vec[i].slice(0,vec[i].length - 1);
	    }
	    resolve(v[v.length - 1]);
	});
	for (var i = v.length - 2; i >= 0; i--) rs = rs.then((function(vi) {
	    return function(vi1) {
		var kbuf = new Uint8Array(32);
		kbuf.set(vi1);
		return crypto.subtle.importKey('raw',kbuf,'AES-CTR',true,['encrypt','decrypt']).then(function(k) {
		    return crypto.subtle.decrypt({name: 'AES-CTR', counter: new Uint8Array(16).fill(0), length: 128},k,vi).then(function(v) {
			return new Uint8Array(v);
		    });
		});
	    };
	})(v[i]));
	return rs;
    },
    scramble: function(vec,b) {
	return libVES.Math.mulv(vec,this.getCv(b));
    },
    explode: function(sc,ct,optns) {
	var self = this;
	return this.toVector(sc).then(function(v) {
	    var bs = self.getBases(ct);
	    var rs = [];
	    for (i = 0; i < bs.length; i++) rs[i] = {
		meta: (function(m) {
		    if (optns) {
			if (optns.t != null) m.t = optns.t;
			if (optns.shadow) for (var k in optns.shadow) m[k] = optns.shadow[k];
		    }
		    return m;
		})({
		    v: self.tag,
		    n: self.size,
		    b: bs[i]
		}),
		value: self.scramble(v,bs[i])
	    }
	    return rs;
	});
    },
    unscramble: function(tokens) {
	var matrix = [];
	var oidx = 0;
	for (var b in tokens) {
	    var row = this.getCv(b);
	    row.push(tokens[b]);
	    matrix.push(row);
	    if (matrix.length >= this.size) break;
	}
	if (matrix.length < this.size) throw new libVES.Error('Internal','Insufficient number of tokens to unscramble');
	return libVES.Math.matrixReduce(matrix).map(function(v,i) {
	    return libVES.Math.div(v[v.length - 1],v[i]);
	});
    },
    implode: function(tokens,then,okfn) {
	var self = this;
	var f = function(offs) {
	    var tks = {};
	    var tidx = 0;
	    var oidx = 0;
	    var more = false;
	    for (var i = 0; i < tokens.length; i++) {
		if (tidx >= self.size) {
		    more = true;
		    break;
		}
		if (tidx >= self.size - offs[oidx]) oidx++;
		else {
		    tks[tokens[i].meta.b] = tokens[i].value;
		    tidx++;
		}
	    }
	    var v = self.unscramble(tks);
	    var rs = self.fromVector(v);
	    if (then) rs = rs.then(then);
	    if (okfn) rs = rs.then(function(sc) {
		for (var i = 0; i < tokens.length; i++) okfn(tokens[i],!!tks[b] && !libVES.Math.cmp(tokens[i].value,self.scramble(v,tokens[i].meta.b)),i);
		return sc;
	    });
	    if (more) {
		var offs2 = offs.slice();
		var jmax = offs.length > 1 ? offs[offs.length - 2] : self.size;
		offs2[offs.length] = 0;
		for (var j = 1; j <= jmax; j++) rs = rs.catch((function(j) {
		    return function() {
			offs2[offs.length - 1] = j;
			return f(offs2);
		    };
		})(j));
	    }
	    return rs;
	};
	return f([0]);
    }
};
libVES.Scramble.algo = {
    'RDX1.2': libVES.Scramble.RDX
};

libVES.Recovery = function(vaultKey, myItems) {
    this.vaultKey = vaultKey;
    if (myItems) this.myItems = Promise.resolve(myItems);
};

libVES.Recovery.prototype = {
    getTokens: function() {
	if (this.tokens) return this.tokens;
	var self = this;
	return this.tokens = self.vaultKey.getType().then(function(t) {
	    switch (t) {
		case 'shadow': case 'recovery': return (self.myItems ? self.myItems : self.vaultKey.getField('vaultItems',{
		    id: true,
		    meta: true,
		    type: true,
		    vaultEntries: {
			encData: true,
			vaultKey: {
			    user: true,
			    type: true
			}
		    }
		},true)).then(function(vis) {
		    var frnds = {};
		    var fn = function() {
			return self.vaultKey.getUser().then(function(my_u) {
			    return my_u.getId().then(function(my_uid) {
				return Promise.all(vis.map(function(vi) {
				    var frnd = {vaultItem: vi};
				    return Promise.all([
					vi.getVaultEntries().then(function(ves) {
					    return Promise.all(ves.map(function(ve) {
						var vk = new libVES.VaultKey(ve.vaultKey,self.vaultKey.VES);
						return vk.getUser().then(function(u) {
						    return u.getId().then(function(uid) {
							if (uid == my_uid) {
							    frnd.assisted = true;
							    return vk.getType().then(function(type) {
								if (type == 'current') frnd.current = true;
							    });
							} else {
							    frnd.user = u;
							    frnds[uid] = frnd;
							}
						    });
						});
					    }));
					}),
					vi.get().then(function(data) {
					    frnd.meta = data.meta;
					    frnd.value = data.value;
					}).catch(function(e) {
					    return vi.getMeta().then(function(meta) {
						frnd.meta = meta;
					    });
					})
				    ]);
				}));
			    });
			}).then(function() {
			    var rs = [];
			    for (var id in frnds) rs.push(frnds[id]);
			    return rs;
			});
		    };
		    return self.vaultKey.trigger ? self.vaultKey.trigger.then(fn) : fn();
		});
		default: throw new libVES.Error('InvalidValue','Recovery info is applicable for key type shadow or recovery');
	    }
	});
    },
    requireOwner: function() {
	return Promise.all([this.vaultKey.getUser(),this.vaultKey.VES.me()]).then(function(usrs) {
	    return Promise.all(usrs.map(function(v,i) {
		return v.getId();
	    })).then(function(uids) {
		if (uids[0] == uids[1]) return true;
		throw new libVES.Error('InvalidValue','Not an owner of the VESrecovery');
	    });
	});
    },
    getFriends: function() {
	return this.getTokens().then(function(tkns) {
	    return tkns.map(function(v,i) {
		return v.user;
	    });
	});
    },
    getOptions: function() {
	return this.getTokens().then(function(tkns) {
	    return tkns.length ? tkns[0].meta : null;
	});
    },
    getFriendInfo: function(user) {
	var self = this;
	return this.getTokens().then(function(tkns) {
	    return Promise.all(tkns.map(function(v,i) {
		return v.user.getId();
	    })).then(function(uids) {
		return user.getId().then(function(uid) {
		    for (var i = 0; i < uids.length; i++) if (uids[i] == uid) return tkns[i];
		    throw new libVES.Error('InvalidValue','Not a friend: ' + uid);
		});
	    });
	});
    },
    getMyToken: function() {
	var self = this;
	return self.vaultKey.VES.me().then(function(me) {
	    return self.getFriendInfo(me);
	});
    },
    getFriendsTotal: function() {
	return this.getTokens().then(function(tkns) {
	    return tkns.length;
	});
    },
    getFriendsRequired: function() {
	return this.getTokens().then(function(tkns) {
	    return tkns[0].meta.n;
	});
    },
    getFriendsAssisted: function() {
	var self = this;
	return this.getTokens().then(function(tkns) {
	    var a = 0;
	    tkns.map(function(t, i) {
		if (t.assisted) a++;
	    });
	    return a;
	});
    },
    getFriendsToGo: function() {
	var self = this;
	return self.getFriendsRequired().then(function(n) {
	    return self.getFriendsAssisted().then(function(a) {
		return a < n ? n - a : 0;
	    });
	});
    },
    _assist: function(assist) {
	var self = this;
	return this.getMyToken().then(function(tkn) {
	    if (!tkn) throw new libVES.Error('InvalidValue','No assistance available');
	    return self.vaultKey.getUser().then(function(user) {
		return self.vaultKey.VES.elevateAuth().catch(function(e) {
		    return undefined;
		}).then(function(optns) {
		    return tkn.vaultItem.shareWith((assist ? [tkn.user,user] : [tkn.user]), undefined, optns).then(function() {
			self.tokens = undefined;
			self.vaultKey.vaultItems = undefined;
			return true;
		    });
		});
	    });
	});
    },
    assist: function() {
	var self = this;
	return self.vaultKey.getType().then(function(t) {
	    return t == 'recovery' ? self._assist(true) : null;
	});
    },
    revoke: function() {
	return this._assist(false);
    },
    unlock: function() {
	var self = this;
	return self.getTokens().then(function(tkns) {
	    var vtkns = [];
	    for (var i = 0; i < tkns.length; i++) if (tkns[i].value != null) vtkns.push(tkns[i]);
	    if (vtkns.length) return libVES.getModule(libVES,['Scramble','algo',vtkns[0].meta.v]).then(function(sc) {
		return new sc(vtkns[0].meta.n).implode(vtkns,function(secret) {
		    return self.vaultKey.unlock(secret);
		});
	    });
	});
    },
    _recover: function() {
	var self = this;
	return self.unlock().then(function() {
	    return self.vaultKey.VES.elevateAuth({authVaultKey: self.vaultKey}).then(function(optns) {
		return self.vaultKey.rekey(optns);
	    });
	});
    },
    recover: function() {
	var self = this;
	if (!this.recovery) this.recovery = this.requireOwner().then(function() {
	    return self._recover();
	});
	return this.recovery;
    }
};

libVES.Delegate = {
    html: '<div style="position:fixed;top:0;left:0;width:100%;height:100%;background:rgba(0,0,0,0.5);display:table;z-index:200000;">'
	+ '<div style="display:table-row;"><div style="display:table-cell;vertical-align:middle;text-align:center;">'
	+ '<div style="min-width:320px;max-width:640px;;background-color:white;margin: auto;padding: 30px;">'
	+ '{$content}'
	+ '</div></div></div></div>',
    htmlDlg: '<p>Use the VESvault popup window to grant access to the App Vault</p>'
	   + '<p class="VESvaultDelegateBlockerMsg" style="color: #bf7f00; font-style:italic;">&nbsp;</p>'
	   + '<p><a class="VESvaultDelegateRetryLnk" href="{$url}" target="VESvaultDelegate" onclick="return !libVES.Delegate.retryPopup(this.href,this)">Click here</a> if you can\'t see VESvault popup window</p>'
	   + '<p><a class="VESvaultDelegateCancelLnk" href="#" onclick="libVES.Delegate.cancel(); return false;">Cancel</a></p>',
    htmlFlw: '<p>Authorizing VES Access</p>'
	   + '<iframe style="width: 100%; height: 48px; border: none;" src="{$url}"></iframe>',
    htmlBlockerMsg: 'Looks like your browser is using a popup blocker...',
    name: 'VESvaultDelegate',
    login: function(VES,challenge,optns) {
	if (this.popup) return this.response || Promise.reject(new libVES.Error('InvalidValue','The delegate popup is already open'));
	if (!challenge) {
	} else try {
	    var info = document.location.search.match(/[\?\&]VESvaultDelegate=([^\&]+)/)[0];
	} catch(e) {}
	this.VES = VES;
	var self = this;
	return this.response = new Promise(function(resolve,reject) {
	    self.reject = reject;
	    self.resolve = resolve;
	    var url = VES.wwwUrl + 'vv/unlock?via=delegate&url=' + encodeURIComponent(document.location.href) + '&domain=' + encodeURIComponent(VES.domain);
	    if (optns) for (var k in optns) if (optns[k] != null) url += '&' + encodeURIComponent(k) + '=' + encodeURIComponent(optns[k]);
	    self.setOrigin(url);
	    self.showOverlay(self.htmlDlg.replace('{$url}',url));
	    self.retryPopupCalled = 0;
	    try {
		document.getElementsByClassName('VESvaultDelegateRetryLnk')[0].onclick = function() {
		    return !libVES.Delegate.retryPopup(this.href,this);
		};
		document.getElementsByClassName('VESvaultDelegateCancelLnk')[0].onclick = function() {
		    libVES.Delegate.cancel();
		    return false;
		};
	    } catch(e) {}
	    if (!self.openPopup(url)) try {
		document.getElementsByClassName('VESvaultDelegateBlockerMsg')[0].innerHTML = self.htmlBlockerMsg;
	    } catch(e) {
		window.alert(self.htmlBlockerMsg);
	    }
	    if (!self.listening) {
		window.addEventListener('message',self.listener.bind(self));
		window.addEventListener('focus',self.chkCancel.bind(self));
		window.addEventListener('beforeunload',self.cancel.bind(self));
		self.listening = true;
	    }
	    window.clearInterval(self.popupInterval);
	    self.popupInterval = window.setInterval(self.chkCancel.bind(self),1000);
	});
    },
    flow: function(VES, start, optns) {
	if (this.response) return this.response;
	this.VES = VES;
	var self = this;
	var unlk = VES.wwwUrl + 'vv/unlock';
	if (start || (!sessionStorage['libVES.flowStarted'] && document.referrer.substr(0, unlk.length).toLowerCase() == unlk.toLowerCase())) {
	    return self.response = new Promise(function(resolve, reject) {
		self.reject = reject;
		self.resolve = resolve;
		var ifrm = VES.wwwUrl + 'vv/flowin?url=' + encodeURIComponent(document.location.href) + '&domain=' + encodeURIComponent(VES.domain);
		if (optns) for (var k in optns) if (optns[k] != null) ifrm += '&' + encodeURIComponent(k) + '=' + encodeURIComponent(optns[k]);
		self.setOrigin(ifrm);
		window.addEventListener('message',self.listener.bind(self));
		self.showOverlay(self.htmlFlw.replace('{$url}', ifrm));
		if (start) delete(sessionStorage['libVES.flowStarted'])
		else sessionStorage['libVES.flowStarted'] = true;
	    });
	} else {
	    return new Promise(function(){});
	}
    },
    showOverlay: function(cont) {
	this.popup = document.createElement('DIV');
	this.popup.innerHTML = this.html.replace('{$content}', cont);
	document.getElementsByTagName('BODY')[0].appendChild(this.popup);
    },
    setOrigin: function(url) {
	this.matchOrigin = (function(m) { return m ? m[0] : document.location.protocol + '//' + document.location.host; })(url.match(/^(https\:\/\/[^\/\?\#]+)/));
    },
    openPopup: function(url) {
	return this.popupWindow = window.open(url,this.name,"width=600,height=600,top=100,left=100");
    },
    retryPopup: function(url,href) {
	this.retryPopupCalled++;
	if (this.popupWindow) try {
	    this.popupWindow.focus();
	    return true;
	} catch(e) {}
	this.openPopup(url);
	return true;
    },
    listener: function(evnt) {
	if (this.popup && evnt.origin == this.matchOrigin) try {
	    var msg = JSON.parse(evnt.data);
	    if (msg.redirect) {
		document.location.replace(msg.redirect);
	    } else {
		this.resolve(this.VES.authorize(msg));
	    }
	    this.close();
	    if (!evnt.source.closed) evnt.source.close();
	} catch(e) {}
    },
    close: function() {
	if (this.popup) {
	    if (this.popupWindow) {
		try {
		    this.popupWindow.close();
		} catch (e) {}
		this.popupWindow = null;
	    }
	    window.clearInterval(this.popupInterval);
	    this.popupInterval = null;
	    this.popup.parentNode.removeChild(this.popup);
	    this.popup = null;
	    return this.response;
	}
    },
    cancel: function() {
	var rs = this.close();
	if (this.response && this.reject) this.reject(new libVES.Error('Aborted','VESvault login is cancelled'));
	return rs;
    },
    chkCancel: function() {
	if (this.popupWindow && this.popupWindow.closed) this.cancel();
    }
};


libVES.Flow = function(name, optns) {
    this.name = name;
    this.origin = document.location.origin;
    this.url = this.origin + '/VESflow';
    if (optns) for (var k in optns) this[k] = optns[k];
    return this;
};

libVES.Flow.algo = {name: 'ECDH', namedCurve: 'P-256'};

libVES.Flow.keyAlgo = function(jwk) {
    if (jwk && jwk.kty == 'EC') return {name: 'ECDH', namedCurve: jwk.crv};
    return null;
};

libVES.Flow.toOrigin = function(url) {
    var a = document.createElement('a');
    a.href = url;
    return a.origin;
};

libVES.Flow.jwkPub = function(jwk) {
    delete(jwk.d);
    delete(jwk.key_ops);
    delete(jwk.ext);
    return jwk;
};

libVES.Flow.parseToken = function(src, callbk) {
    src = src.replace(/\#VESflow\.(\w+)=([^\#\&\=]+)/g, function(s, k, v) {
	var tk;
	try {
	    tk = JSON.parse(decodeURIComponent(v));
	    if (callbk(decodeURIComponent(k), tk)) return '';
	} catch(e) {
	    console.log(e);
	}
	return s;
    });
    return src;
};

libVES.Flow.logout = function() {
    for (var k in sessionStorage) if (k.match(/^VESflow\b/)) delete(sessionStorage[k]);
};

libVES.Flow.prototype.source = function(src) {
    if (src != null) history.replaceState(null, '', src);
    return document.location.href;
};

libVES.Flow.prototype.privKey = function(algo) {
    if (!algo) algo = libVES.Flow.algo;
    var a = JSON.stringify(algo);
    return Promise.resolve(sessionStorage['VESflow' + a]).then(function(k) {
	if (!k) throw null;
	return crypto.subtle.importKey('jwk', JSON.parse(k), algo, true, ['deriveBits', 'deriveKey']);
    }).catch(function(e) {
	return crypto.subtle.generateKey(algo, true, ['deriveBits', 'deriveKey']).then(function(pair) {
	    return crypto.subtle.exportKey('jwk', pair.privateKey).then(function(jwk) {
		sessionStorage['VESflow' + a] = JSON.stringify(jwk);
		return pair.privateKey;
	    });
	});
    });
};

libVES.Flow.prototype.pubJWK = function(org) {
    return ((!org || org == this.origin) ? this.privKey().then(function(priv) {
	return crypto.subtle.exportKey('jwk', priv).then(function(jwk) {
	    return libVES.Flow.jwkPub(jwk);
	});
    }) : Promise.resolve(sessionStorage['VESflow|' + org]).then(function(k) {
	if (!k) throw new libVES.Error('NotFound', 'No public key stored for ' + org);
	return JSON.parse(k);
    }));
};

libVES.Flow.prototype.pubKey = function(org) {
    return this.pubJWK(org).then(function(jwk) {
	return crypto.subtle.importKey('jwk', jwk, libVES.Flow.keyAlgo(jwk), true, []);
    });
};

libVES.Flow.prototype.cipher = function(url) {
    var self = this;
    return self.pubKey(libVES.Flow.toOrigin(url)).then(function(pub) {
	return self.privKey(pub.algorithm).then(function(priv) {
	    return crypto.subtle.deriveBits({name: 'ECDH', public: pub}, priv, 8 * libVES.Algo.ECDH.curveBytes[pub.algorithm.namedCurve]);
	});
    }).then(function(raw) {
	return crypto.subtle.digest({name: 'SHA-256'}, raw).then(function(buf) {
	    return crypto.subtle.importKey('raw', buf, 'AES-GCM', false, ['encrypt', 'decrypt']);
	});
    });
};

libVES.Flow.prototype.encrypt = function(url, val) {
    var self = this;
    return Promise.resolve(val || this.value()).then(function(val) {
	if (val == null) throw new libVES.Error('NoData', 'Empty value');
	return self.cipher(url).then(function(ci) {
	    var iv = new Uint8Array(12);
	    crypto.getRandomValues(iv);
	    return crypto.subtle.encrypt({name: 'AES-GCM', iv: iv.buffer}, ci, libVES.Util.StringToByteArray(JSON.stringify(val))).then(function(ctext) {
		return libVES.Util.ByteArrayToB64W(iv) + '.' + libVES.Util.ByteArrayToB64W(ctext);
	    });
	});
    });
};

libVES.Flow.prototype.token = function(url) {
    var self = this;
    return this.pubJWK().then(function(jwk) {
	return {url: self.url, key: jwk};
    }).then(function(tk) {
	return self.encrypt(url).then(function(ctext) {
	    tk.enc = ctext;
	    self.erase(url);
	    self.sent = true;
	    return tk;
	}).catch(function(e) {
	    self.store(url).catch(function(e) {});
	    return tk;
	});
    });
};

libVES.Flow.prototype.addToken = function(url) {
    var self = this;
    return self.token(url).then(function(tk) {
	return url + '#VESflow.' + self.name + '=' + encodeURIComponent(JSON.stringify(tk));
    });
};

libVES.Flow.prototype.decrypt = function(ctext, url) {
    var self = this;
    var c = ctext.split('.');
    return self.cipher(url).then(function(ci) {
	return crypto.subtle.decrypt({name: 'AES-GCM', iv: libVES.Util.B64ToByteArray(c[0])}, ci, libVES.Util.B64ToByteArray(c[1])).then(function(ptext) {
	    return JSON.parse(libVES.Util.ByteArrayToString(ptext));
	});
    });
};

libVES.Flow.prototype.value = function() {
    var self = this;
    return Promise.resolve(history.state && history.state['VESflow.' + self.name]).then(function(ctext) {
	return self.decrypt(ctext);
    });
};

libVES.Flow.prototype.setValue = function(val) {
    var self = this;
    return Promise.resolve(val? self.encrypt(null, val) : null).then(function(ctext) {
	var s = {};
	if (history.state) for (var k in history.state) s[k] = history.state[k];
	if (ctext) s['VESflow.' + self.name] = ctext;
	else delete(s['VESflow.' + self.name]);
	history.replaceState(s, '');
	return val;
    });
};

libVES.Flow.prototype.click = function(a) {
    var self = this;
    (a.origin == this.origin ? self.store() : self.addToken(a.href).then(function(url) {
	a.href = url;
    })).catch(function(e) {
	if (!e || e.code != 'NotFound') throw e;
    }).then(function() {
	document.location.href = a.href;
    }).catch(function(e) {
	console.log(e);
    });
    return false;
};

libVES.Flow.prototype.reload = function(e) {
    if (!this.sent) this.store().catch(function(e) {});
};

libVES.Flow.prototype.recv = function(dval) {
    var self = this;
    return Promise.resolve(self.source()).then(function(src) {
	var tk = null;
	src = libVES.Flow.parseToken(src, function(k, t) {
	    if (k != self.name) return false;
	    tk = t;
	    return true;
	});
	if (tk == null) throw new libVES.Error('NotFound', 'No valid token in the source url');
	return Promise.resolve(self.source(src)).then(function() {
	    if (dval) for (var k in dval) if (tk[k] === undefined) tk[k] = dval[k];
	    return self.recvToken(tk);
	});
    });
};

libVES.Flow.prototype.recvToken = function(tk) {
    var org = libVES.Flow.toOrigin(tk.url);
    if (org != this.origin && tk.key) sessionStorage['VESflow|' + org] = JSON.stringify(tk.key);
    if (!tk.enc) return Promise.reject(new libVES.Error('Incomplete', 'Missing encrypted data', {url: tk.url}));
    return this.decrypt(tk.enc, org);
};

libVES.Flow.prototype.get = function() {
    var self = this;
    return self.fetch().catch(function(e) {
	if (!e || e.code != 'NotFound') console.log(e);
	return null;
    }).then(function(data) {
	return self.recv().catch(function(e) {
	    if (e && e.code == 'Incomplete') {
		self.url = document.location.href;
		return self.addToken(e.url).then(function(url) {
		    document.location.replace(url);
		    throw new libVES.Error('Redirect', 'Completing the key exchange');
		});
	    }
	    if (!data) throw e;
	    return data;
	});
    });
};

libVES.Flow.prototype.store = function(url) {
    if (history.state && history.state['VESflow.' + this.name]) {
	this.sent = 1;
	return Promise.resolve(sessionStorage['VESflow.' + this.name + '>' + libVES.Flow.toOrigin(url)] = history.state['VESflow.' + this.name]);
    }
    return Promise.reject(new libVES.Error('NotFound', 'Value is not set'));
};

libVES.Flow.prototype.fetch = function(url) {
    var self = this;
    return Promise.resolve(sessionStorage['VESflow.' + self.name + '>' + libVES.Flow.toOrigin(url)]).then(function(ctext) {
	if (!ctext) throw new libVES.Error('NotFound', 'Not in the session storage');
	return self.decrypt(ctext).then(function(data) {
	    self.erase(url);
	    return data;
	});
    });
};

libVES.Flow.prototype.erase = function(url) {
    delete(sessionStorage['VESflow.' + this.name + '>' + libVES.Flow.toOrigin(url)]);
};

libVES.Flow.prototype.logout = function() {
    this.setValue(undefined);
    return libVES.Flow.logout();
};


libVES.Auth = function(optns) {
    for (var k in optns) this[k] = optns[k];
    return this;
};

libVES.Auth.prototype.getToken = function() {
    if (this.vaultItem) return Promise.all([this.vaultItem.getId(), this.vaultItem.getField('verifyToken')]).then(function(flds) {
	return 'vaultItem.' + flds[0] + '.' + flds[1];
    });
    else if (this.vaultKey) return Promise.all([this.vaultKey.getId(), this.vaultItem.getSessionToken()]).then(function(flds) {
	return 'vaultKey.' + flds[0] + '.' + flds[1];
    });
    else if ((VES = this.VES) && VES.token) return this.VES.getVaultKey().then(function(vkey) {
	return vkey.getId().then(function(id) {
	    return 'vaultKey.' + id + '.' + VES.token;
	});
    });
    return Promise.reject({code: 'VESauth', message: 'No object to generate a VESauth token for'});
};

libVES.Auth.prototype.getData = function(url, type) {
    return this.getToken().then(function(tkn) {
	return new Promise(function(resolve, reject) {
	    var xhr = new XMLHttpRequest();
	    xhr.open('GET', url);
	    if (type) xhr.responseType = type;
	    xhr.onreadystatechange = function() {
		switch(xhr.readyState) {
		    case 4:
			if (xhr.status == 200) resolve(xhr.response);
			else reject({code: xhr.status, message: xhr.statusText});
		}
	    };
	    xhr.setRequestHeader('X-VES-Authorization', tkn);
	    xhr.send();
	});
    });
};

libVES.Auth.prototype.getJSON = function(url) {
    return this.getData(url, 'json').then(function(json) {
	var a = document.createElement('a');
	a.href = url;
	if (!a.hash) return json;
	var path = a.hash.split(/\//);
	for (var i = 1; i < path.length; i++) if (path[i]) {
	    if (json instanceof Array) json = json[Number(path[i])];
	    else if (json instanceof Object) json = json[path[i]];
	    else return null;
	}
	return json;
    });
};

libVES.Auth.prototype.auth = function(token) {
    return this.verify(token, true);
};

libVES.Auth.prototype.verify = function(token, auth) {
    var tk = token.match(/^(\w*)\.(\d*)\.(.*)$/);
    if (!tk) return Promise.reject(new libVES.Error('VESauth', 'Invalid VESauth token'));
    var ves = new libVES({domain: (this.domain || '*'), token: tk[3]});
    if (tk[1] == 'vaultItem') {
	if (auth) return Promise.reject(new libVES.Error('VESauth', 'Verify token cannot be used for authentication'));
	return Promise.resolve(this.vaultItem ? this.vaultItem.getId() : tk[2]).then(function(id) {
	    var vi = new libVES.VaultItem({id: id}, ves);
	    return vi.getFile().then(function(file) {
		return vi.getField('deleted').then(function(f) {
		    if (f) throw new libVES.Error('VESauth', 'Verify token expired');
		    return file.getCreator();
		});
	    });
	});
    }
    if (tk[1] != 'vaultKey') return Promise.reject(new libVES.Error('VESauth', 'Unknown token type'));
    if (this.vaultItem) return this.vaultItem.getId().then(function(id) {
	var vi = new libVES.VaultItem({id: id}, ves);
	return vi.getVaultEntries().then(function(vents) {
	    if (!auth && vents && vents.length) return ves.me();
	    if (vents && tk[2]) for (var i = 0; i < vents.length; i++) if (vents[i].vaultKey && Number(vents[i].vaultKey.id) == Number(tk[2])) return ves.me();
	    throw new libVES.Error('VESauth', 'ACL not authorized');
	});
    });
    if (!auth) return ves.me();
    return Promise.resolve(this.vaultKey ? this.vaultKey.getId() : tk[2]).then(function(id) {
	return (new libVES.VaultKey({id: id}, ves)).getExternals().then(function(ext) {
	    if (!ext || !ext[0]) throw new libVES.Error('VESauth', 'Bad vaultKey for VESauth');
	    return ext[0].getDomain().then(function(dom) {
		return ext[0].getExternalId().then(function(extid) {
		    if (!extid || !extid.match(/^[^\!\@]+\@[^\!\@]+$/)) throw new libVES.Error('VESauth', 'Bad vaultKey for VESauth');
		    if (dom && dom.toLowerCase() == ves.domain.toLowerCase()) return ves.me();
		    throw new libVES.Error('VESuath', 'VES domain does not match');
		});
	    });
	});
    });
};


libVES.Algo.OQS = {
    tag: 'OQS',
    wasm: function() {
	if (!this.wasmP) this.wasmP = (typeof(WasmOQS) == 'function' ? Promise.resolve()
	    : libVES.Util.loadWasm(WasmOQSinit.baseUrl + 'WasmOQS.js')
	).then(function() {
	    return WasmOQS(WasmOQSinit);
	});
	return this.wasmP;
    },
    checkLimits: function(k) {
	return k.pub.byteLength + k.priv.byteLength <= libVES.maxKeyLen && k.ctextlen + 64 <= libVES.maxEncDataLen;
    },
    decrypt: function(k,buf) {
	return this.wasm().then(function(wasm) {
	    var b = new Uint8Array(buf);
	    if (b.byteLength <= k.ctextlen) throw new libVES.Error('InvalidValue', 'Ciphertext is too short');
	    var secret = wasm.decaps(k, b.slice(0, k.ctextlen));
	    if (!secret) throw new libVES.Error('InvalidValue', 'OQS decryption error');
	    return libVES.Algo.OQS.cipher(secret).then(function(ci) {
		return ci.decrypt(b.slice(k.ctextlen), true).catch(function(e) {
		    if (e instanceof libVES.Error) throw e;
		    throw new libVES.Error('InvalidValue', 'Error decrypting ECIES payload: Invalid ciphertext?',{error:e});
		});
	    }).catch(function(e) {
		if (e instanceof libVES.Error) throw e;
		throw new libVES.Error('InvalidValue', 'Error negotiating ECIES cipher key: Invalid ciphertext?',{error:e});
	    });
	})
    },
    encrypt: function(k, buf) {
	return this.wasm().then(function(wasm) {
	    var ctext = wasm.encaps(k);
	    if (!ctext) throw new libVES.Error('InvalidValue', 'Error generating OQS secret');
	    return libVES.Algo.OQS.cipher(wasm.decaps(k)).then(function(ci) {
		return ci.encrypt(buf, true).then(function(ctext2) {
		    var buf = new Uint8Array(ctext.byteLength + ctext2.byteLength);
		    buf.set(new Uint8Array(ctext),0);
		    buf.set(new Uint8Array(ctext2),ctext.byteLength);
		    return buf;
		});
	    }).catch(function(e) {
		if (e instanceof libVES.Error) throw e;
		throw new libVES.Error('InvalidValue', 'Error generating ECIES cipher',{error:e});
	    });
	});
    },
    cipher: function(secret) {
	return libVES.getModule(libVES.Cipher,'AES256GCMp').then(function(m) {
	    return crypto.subtle.digest({name:'SHA-384'}, secret).then(function(buf) {
		return new m(new Uint8Array(buf).slice(0, m.prototype.keySize + m.prototype.ivSize));
	    });
	});
    },
    import: function(data,optns) {
	return libVES.Util.PEM.import(data,optns);
    },
    export: function(data, optns) {
	if (!data || !data.ptr) throw new libVES.Error('Internal', "Unknown type of key object");
	var oid = [new libVES.Util.OID(libVES.Algo.OQS.OID), libVES.Util.StringToByteArray(data.algo)];
	return libVES.Util.Key.export(data.pub, data.priv, oid, optns);
    },
    generate: function(optns) {
	var algo = optns ? optns.oqsAlgo : null;
	if (!algo) algo = this.defaultAlgo;
	return this.wasm().then(function(wasm) {
	    var k = wasm.init(algo, true);
	    if (!k) throw new libVES.Error('InvalidValue', 'OQS key init failed');
	    if (!libVES.Algo.OQS.checkLimits(k)) throw wasm.free(k), new libVES.Error('Internal', 'OQS key is too large for VES');
	    if (!wasm.generate(k)) throw wasm.free(k), new libVES.Error('Internal', 'OQS generate failed');
	    var pub = new wasm.Key(k.algo);
	    for (var i in k) if (i != 'priv') pub[i] = k[i];
	    return {privateKey: k, publicKey: pub};
	});
    },
    getKeyOptions: function(key) {
	return {oqsAlgo: k.algo};
    },
    getMethods: function() {
	return this.wasm().then(function(wasm) {
	    return wasm.getalgos().filter(function(algo) {
		var k = wasm.init(algo, true);
		if (!k) return false;
		var f = libVES.Algo.OQS.checkLimits(k);
		wasm.free(k);
		return f;
	    }).map(function(algo) {
		return {oqsAlgo: algo};
	    });
	});
    },
    Util: {
	import: function(args,chain,optns) {
	    var algo = libVES.Util.ByteArrayToString(args);
	    return chain('container').then(function(der) {
		return libVES.Util.Key.fromDER(der, function(pub, priv, pubBits) {
		    return libVES.Algo.OQS.wasm().then(function(wasm) {
			var k = wasm.init(algo, !!priv);
			if (!k) throw new libVES.Error('InvalidValue', 'OQS key init failed (bad algo?)');
			var s = function(dst, src) {
			    if (!dst || !src || dst.byteLength != src.byteLength) throw wasm.free(k), new libVES.Error('InvalidValue', 'Incorrect OQS key size');
			    dst.set(src, 0);
			};
			if (priv) s(k.priv, priv);
			s(k.pub, pub);
			return k;
		    });
		});
	    });
	}
    },
    OID: '1.3.6.1.4.1.53675.3.5',
    defaultAlgo: 'Kyber768'
};
libVES.Algo.all.push('OQS');

var WasmOQSinit = {
    Key: function(algo) {
	this.algo = algo;
    },
    init: function(algo, fpriv) {
	var a = new Uint8Array(libVES.Util.StringToByteArray(algo));
	var arg1 = new Uint8Array(this.asm.memory.buffer, 16, 48);
	arg1.set(a);
	arg1.set([0], a.byteLength);
	var ptr = this._WasmOQS_new(arg1.byteOffset, fpriv);
	if (!ptr) return null;
	var key = new this.Key(algo);
	key.ptr = ptr;
	var priv = this._WasmOQS_priv(ptr);
	if (priv) key.priv = new Uint8Array(this.asm.memory.buffer, priv, this._WasmOQS_privlen(ptr));
	key.pub = new Uint8Array(this.asm.memory.buffer, this._WasmOQS_pub(ptr), this._WasmOQS_publen(ptr));
	key.secretlen = this._WasmOQS_decaps(key.ptr, null, null);
	key.ctextlen = this._WasmOQS_encaps(key.ptr, null, null);
	return key;
    },
    generate: function(key) {
	return this._WasmOQS_generate(key.ptr) >= 0;
    },
    encaps: function(key) {
	var arg1 = new Uint8Array(this.asm.memory.buffer, this._WasmOQS_secretbuf, this._WasmOQS_decaps(key.ptr, null, null));
	var arg2 = new Uint8Array(this.asm.memory.buffer, this._WasmOQS_ctextbuf, this._WasmOQS_encaps(key.ptr, null, null));
	if (this._WasmOQS_encaps(key.ptr, arg2.byteOffset, arg1.byteOffset) <= 0) return null;
	key.secret = new Uint8Array(arg1.byteLength);
	key.secret.set(arg1);
	var ctext = new Uint8Array(arg2.byteLength);
	ctext.set(arg2);
	return ctext;
    },
    decaps: function(key, ctext) {
	var arg1 = new Uint8Array(this.asm.memory.buffer, this._WasmOQS_secretbuf, this._WasmOQS_decaps(key.ptr, null, null));
	if (ctext) {
	    if (ctext.byteLength != this._WasmOQS_encaps(key.ptr, null, null)) return null;
	    var arg2 = new Uint8Array(this.asm.memory.buffer, this._WasmOQS_ctextbuf, ctext.byteLength);
	    arg2.set(ctext);
	    if (this._WasmOQS_decaps(key.ptr, arg1.byteOffset, arg2.byteOffset) <= 0) return null;
	}
	var secret = new Uint8Array(arg1.byteLength);
	secret.set(arg1);
	return secret;
    },
    free: function(key) {
	this._WasmOQS_free(key.ptr);
	delete(key.ptr);
    },
    getalgos: function() {
	var algos = [];
	var a;
	for (var i = 0; a = this._WasmOQS_enumalgo(i); i++) {
	    var s = new Uint8Array(this.asm.memory.buffer, a);
	    var p = s.indexOf(0);
	    if (!p) continue;
	    algos.push(libVES.Util.ByteArrayToString(s.slice(0, p)));
	}
	return algos;
    },
    test: function(algo) {
	var k = this.init(algo, 1);
	this.generate(k);
	var ctext = this.encaps(k);
	var secret = this.decaps(k);
	console.log(secret);
	var secret2 = this.decaps(k, ctext);
	console.log(secret2);
    },
    locateFile: function(file) {
	return this.baseUrl + file;
    },
    baseUrl: 'https://ves.host/pub/'
};



libVES.Watch = function(obj, optns) {
    if (obj instanceof libVES) obj = obj.domain ? (obj.externalId ? obj.getVaultKey() : new libVES.Domain(null, obj)) : obj.me();
    var self = this;
    this.object = Promise.resolve(obj).then(function(o) {
	if (o instanceof libVES.VaultKey) {
	    self.uri = 'vaultKeys/';
	} else if (o instanceof libVES.VaultItem) {
	    self.uri = 'vaultItems/';
	} else if (o instanceof libVES.User) {
	    self.uri = 'users/';
	} else if (o instanceof libVES.Domain) {
	    self.uri = 'domains/';
	} else {
	    throw new libVES.Error('Internal', 'Watchable object expected');
	}
	self.VES = o.VES;
	return o;
    });
    if (optns) for (var k in optns) this[k] = optns[k];
};

libVES.Watch.prototype = {
    fields: {vaultKey: {id: true, user: true, externals: true, creator: true}, vaultItem: {id: true, vaultKey: true, file: {externals: true, creator: true}}, session: {id: true, userAgent: true, remote: true, createdAt: true, expiresAt: true}},
    state: 'idle',
    timeout: 900,
    getEvents: function(start, ct, poll) {
	var self = this;
	return self.object.then(function(o) {
	    return o.getId().then(function(id) {
		return (poll ? self.VES.pollUrl : '') + self.uri + id
		    + '?fields=events(' + self.VES.uriListFields(self.fields) + ')%5B' + start + (ct >= 0 ? '%2B' : '') + ct + '%5D'
		    + (poll ? '&poll=' + self.timeout : '');
	    });
	}).then(function(url) {
	    return self.VES.get(url);
	}).then(function(res) {
	    if (!res.events) return null;
	    return res.events.map(function(e, i) {
		return new libVES.Event(e, self.VES);
	    });
	});
    },
    setState: function(st) {
	if (this.statefn) try {
	    this.statefn(st, this);
	} catch(e) {}
	this.state = st;
    },
    setBusy: function(busy) {
	this.busy = busy;
	if (!busy && this.state == 'stop') this.setState('idle');
    },
    start: function(pos) {
	if (!this.eventfn) throw new libVES.Error('InvalidValue', 'eventfn is not set');
	if (this.busy) throw new libVES.Error('Internal', 'Watch instance is busy');
	var self = this;
	self.setBusy(true);
	this.setState('read');
	if (pos < 0) {
	    this.next = 0;
	    return this.getEvents(0, pos).then(function(evs) {
		self.setBusy(false);
		return self.received(evs.reverse());
	    });
	} else {
	    this.next = pos;
	    return this.getEvents(pos, 0).then(function(evs) {
		self.setBusy(false);
		return self.received(evs);
	    });
	}
    },
    active: function() {
	return this.state == 'read' || this.state == 'poll';
    },
    received: function(evs) {
	var self = this;
	var p;
	if (evs && evs.length) p = evs.reduce(function(v, e) {
	    if (self.active()) try {
		self.eventfn(e, self);
	    } catch(e) {
		console.log(e);
	    }
	    return e;
	}, null).getId().then(function(id) {
	    self.next = id + 1;
	});
	else {
	    p = Promise.resolve(null);
	    if (this.state == 'read') this.setState('poll');
	}
	return p.then(function() {
	    return self.request();
	});
    },
    request: function() {
	if (this.busy || !this.active()) return;
	var self = this;
	self.setBusy(true);
	return self.getEvents(self.next, 0, (self.state == 'poll')).then(function(evs) {
	    self.setBusy(false);
	    self.received(evs);
	}).catch(function(e) {
	    self.setBusy(false);
	    if (self.errorfn) try {
		self.errorfn(e, self);
	    } catch(e) {
		console.log(e);
	    }
	    return self.retry();
	});
    },
    retry: function() {
	var self = this;
	return new Promise(function(resolve, reject) {
	    window.setTimeout(function() {
		resolve(self.request());
	    }, 1000);
	});
    },
    stop: function() {
	this.setState(this.busy ? 'stop' : 'idle');
    }
};
libVES.version = '2.16';

module.exports = libVES;
